#***************************************************************************
#*   (c) Juergen Riegel (juergen.riegel@web.de) 2003                       *
#*                                                                         *
#*   This file is part of the FreeCAD CAx development system.              *
#*                                                                         *
#*   This program is free software; you can redistribute it and/or modify  *
#*   it under the terms of the GNU Lesser General Public License (LGPL)    *
#*   as published by the Free Software Foundation; either version 2 of     *
#*   the License, or (at your option) any later version.                   *
#*   for detail see the LICENCE text file.                                 *
#*                                                                         *
#*   FreeCAD is distributed in the hope that it will be useful,            *
#*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
#*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
#*   GNU Library General Public License for more details.                  *
#*                                                                         *
#*   You should have received a copy of the GNU Library General Public     *
#*   License along with FreeCAD; if not, write to the Free Software        *
#*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
#*   USA                                                                   *
#*                                                                         *
#*   Juergen Riegel 2003                                                   *
#***************************************************************************/

import FreeCAD, os, unittest, tempfile
import math

#---------------------------------------------------------------------------
# define the functions to test the FreeCAD Document code
#---------------------------------------------------------------------------


class DocumentBasicCases(unittest.TestCase):
  def setUp(self):
    self.Doc = FreeCAD.newDocument("CreateTest")

  def testCreateDestroy(self):
    #FIXME: Causes somehow a ref count error but it's _not_ FreeCAD.getDocument()!!!
    #If we remove the whole method no error appears.
    self.failUnless(FreeCAD.getDocument("CreateTest")!= None,"Creating Document failed")

  def testAddition(self):
    # Cannot write a real test case for that but when debugging the
    # C-code there shouldn't be a memory leak (see rev. 1814)
    self.Doc.openTransaction("Add")
    L1 = self.Doc.addObject("App::FeatureTest","Label")
    self.Doc.commitTransaction()
    self.Doc.undo()

  def testAddRemoveUndo(self):
    # Bug #0000525
    self.Doc.openTransaction("Add")
    obj=self.Doc.addObject("App::FeatureTest","Label")
    self.Doc.commitTransaction()
    self.Doc.removeObject(obj.Name)
    self.Doc.undo()
    self.Doc.undo()

  def testNoRecompute(self):
    L1 = self.Doc.addObject("App::FeatureTest","Label")
    self.Doc.recompute()
    L1.TypeNoRecompute = 2
    execcount = L1.ExecCount
    objectcount = self.Doc.recompute()
    self.assertEqual(objectcount, 0)
    self.assertEqual(L1.ExecCount, execcount)

  def testNoRecomputeParent(self):
    L1 = self.Doc.addObject("App::FeatureTest","Child")
    L2 = self.Doc.addObject("App::FeatureTest","Parent")
    L2.Source1 = L1
    self.Doc.recompute()
    L1.TypeNoRecompute = 2
    countChild = L1.ExecCount
    countParent = L2.ExecCount
    objectcount = self.Doc.recompute()
    self.assertEqual(objectcount, 1)
    self.assertEqual(L1.ExecCount, countChild)
    self.assertEqual(L2.ExecCount, countParent+1)

    L1.touch('')
    countChild = L1.ExecCount
    countParent = L2.ExecCount
    objectcount = self.Doc.recompute()
    self.assertEqual(objectcount, 1)
    self.assertEqual(L1.ExecCount, countChild)
    self.assertEqual(L2.ExecCount, countParent+1)

    L1.enforceRecompute()
    countChild = L1.ExecCount
    countParent = L2.ExecCount
    objectcount = self.Doc.recompute()
    self.assertEqual(objectcount, 2)
    self.assertEqual(L1.ExecCount, countChild+1)
    self.assertEqual(L2.ExecCount, countParent+1)

  def testAbortTransaction(self):
    self.Doc.openTransaction("Add")
    obj=self.Doc.addObject("App::FeatureTest","Label")
    self.Doc.abortTransaction()
    TempPath = tempfile.gettempdir()
    SaveName = TempPath + os.sep + "SaveRestoreTests.FCStd"
    self.Doc.saveAs(SaveName)

  def testRemoval(self):
    # Cannot write a real test case for that but when debugging the
    # C-code there shouldn't be a memory leak (see rev. 1814)
    self.Doc.openTransaction("Add")
    L1 = self.Doc.addObject("App::FeatureTest","Label")
    self.Doc.commitTransaction()
    self.Doc.openTransaction("Rem")
    L1 = self.Doc.removeObject("Label")
    self.Doc.commitTransaction()

  def testObjects(self):
    L1 = self.Doc.addObject("App::FeatureTest","Label_1")
    #call members to check for errors in ref counting
    self.Doc.ActiveObject
    self.Doc.Objects
    self.Doc.UndoMode
    self.Doc.UndoRedoMemSize
    self.Doc.UndoCount
    # test read only mechanismus
    try:
      self.Doc.UndoCount = 3
    except:
      FreeCAD.Console.PrintLog("   exception thrown, OK\n")
    else:
      self.fail("no exception thrown")
    self.Doc.RedoCount
    self.Doc.UndoNames
    self.Doc.RedoNames
    self.Doc.recompute()
    self.failUnless(L1.Integer == 4711)
    self.failUnless(L1.Float-47.11<0.001)
    self.failUnless(L1.Bool    == True)
    self.failUnless(L1.String  == "4711")
    #temporarily not checked because of strange behavior of boost::filesystem JR
    #self.failUnless(L1.Path  == "c:/temp")
    self.failUnless(float(L1.Angle)-3.0<0.001)
    self.failUnless(float(L1.Distance)-47.11<0.001)

    # test basic property stuff
    self.failUnless(not L1.getDocumentationOfProperty("Source1") == "")
    self.failUnless(L1.getGroupOfProperty("Source1") == "Feature Test")
    self.failUnless(L1.getTypeOfProperty("Source1") == [])


    # test the constraint types ( both are constraint to percent range)
    self.failUnless(L1.ConstraintInt == 5)
    self.failUnless(L1.ConstraintFloat-5.0<0.001)
    L1.ConstraintInt = 500
    L1.ConstraintFloat = 500.0
    self.failUnless(L1.ConstraintInt == 100)
    self.failUnless(L1.ConstraintFloat - 100.0 < 0.001)
    L1.ConstraintInt = -500
    L1.ConstraintFloat = -500.0
    self.failUnless(L1.ConstraintInt == 0)
    self.failUnless(L1.ConstraintFloat - 0.0 < 0.001)

    # test enum property
    # in App::FeatureTest the current value is set to 4
    self.failUnless(L1.Enum  == "Four")
    L1.Enum = "Three"
    self.failUnless(L1.Enum  == "Three",     "Different value to 'Three'")
    L1.Enum = 2
    self.failUnless(L1.Enum  == "Two",     "Different value to 'Two'")
    try:
      L1.Enum = "SurelyNotInThere!"
    except:
      FreeCAD.Console.PrintLog("   exception thrown, OK\n")
    else:
      self.fail("no exception thrown")

    #self.failUnless(L1.IntegerList  == [4711]   )
    #f = L1.FloatList
    #self.failUnless(f -47.11<0.001    )
    #self.failUnless(L1.Matrix  == [1.0,2.0,3.0,4.0,5.0,6.0,7.0,8.0,9.0,10.0,11.0,12.0,13.0,14.0,15.0,16.0] )
    #self.failUnless(L1.Vector  == [1.0,2.0,3.0])

    self.failUnless(L1.Label== "Label_1","Invalid object name")
    L1.Label="Label_2"
    self.Doc.recompute()
    self.failUnless(L1.Label== "Label_2","Invalid object name")
    self.Doc.removeObject("Label_1")

  def testEnum(self):
    enumeration_choices = ["one", "two"]
    obj = self.Doc.addObject("App::FeaturePython","Label_2")
    obj.addProperty("App::PropertyEnumeration", "myEnumeration", "Enum", "mytest")
    with self.assertRaises(ValueError):
      obj.myEnumeration = enumeration_choices[0]

  def testMem(self):
    self.Doc.MemSize

  def testDuplicateLinks(self):
    obj = self.Doc.addObject("App::FeatureTest","obj")
    grp = self.Doc.addObject("App::DocumentObjectGroup","group")
    grp.Group = [obj,obj]
    self.Doc.removeObject(obj.Name)
    self.assertListEqual(grp.Group, [])

  def testPlacementList(self):
    obj = self.Doc.addObject("App::FeaturePython","Label")
    obj.addProperty("App::PropertyPlacementList", "PlmList")
    plm = FreeCAD.Placement()
    plm.Base = (1,2,3)
    plm.Rotation = (0,0,1,0)
    obj.PlmList = [plm]
    cpy = self.Doc.copyObject(obj)
    self.assertListEqual(obj.PlmList, cpy.PlmList)

  def testAddRemove(self):
    L1 = self.Doc.addObject("App::FeatureTest","Label_1")
    # must delete object
    self.Doc.removeObject(L1.Name)
    try:
      L1.Name
    except:
      self.failUnless(True)
    else:
      self.failUnless(False)
    del L1

    # What do we expect here?
    self.Doc.openTransaction("AddRemove")
    L2 = self.Doc.addObject("App::FeatureTest","Label_2")
    self.Doc.removeObject(L2.Name)
    self.Doc.commitTransaction()
    self.Doc.undo()
    try:
      L2.Name
    except:
      self.failUnless(True)
    else:
      self.failUnless(False)
    del L2

  def testExtensions(self):
    #we try to create a normal python object and add an extension to it
    obj = self.Doc.addObject("App::DocumentObject", "Extension_1")
    grp = self.Doc.addObject("App::DocumentObject", "Extension_2")
    #we should have all methods we need to handle extensions
    try:
      self.failUnless(not grp.hasExtension("App::GroupExtensionPython"))
      grp.addExtension("App::GroupExtensionPython", self)
      self.failUnless(grp.hasExtension("App::GroupExtension"))
      self.failUnless(grp.hasExtension("App::GroupExtensionPython"))
      grp.addObject(obj)
      self.failUnless(len(grp.Group) == 1)
      self.failUnless(grp.Group[0] == obj)
    except:
      self.failUnless(False)

    #test if the method override works
    class SpecialGroup():
        def allowObject(self, obj):
            return False;

    callback = SpecialGroup()
    grp2 = self.Doc.addObject("App::DocumentObject", "Extension_3")
    grp2.addExtension("App::GroupExtensionPython", callback)

    try:
      self.failUnless(grp2.hasExtension("App::GroupExtension"))
      grp2.addObject(obj)
      self.failUnless(len(grp2.Group) == 0)
    except:
      self.failUnless(True)

    self.Doc.removeObject(grp.Name)
    self.Doc.removeObject(grp2.Name)
    self.Doc.removeObject(obj.Name)
    del obj
    del grp
    del grp2

  def testExtensionBug0002785(self):

        class MyExtension():
            def __init__(self, obj):
                obj.addExtension("App::GroupExtensionPython", self)

        obj = self.Doc.addObject("App::DocumentObject", "myObj")
        MyExtension(obj)
        self.failUnless(obj.hasExtension("App::GroupExtension"))
        self.failUnless(obj.hasExtension("App::GroupExtensionPython"))
        self.Doc.removeObject(obj.Name)
        del obj

  def testExtensionGroup(self):
    obj = self.Doc.addObject("App::DocumentObject", "Obj")
    grp = self.Doc.addObject("App::FeaturePython", "Extension_2")
    grp.addExtension("App::GroupExtensionPython", None)
    grp.Group = [obj]
    self.assertTrue(obj in grp.Group)

  def testExtensionBugViewProvider(self):

    class Layer():
      def __init__(self, obj):
        obj.addExtension("App::GroupExtensionPython", self)

    class LayerViewProvider():
      def __init__(self, obj):
        obj.addExtension("Gui::ViewProviderGroupExtensionPython", self)
        obj.Proxy = self

    obj = self.Doc.addObject("App::FeaturePython","Layer")
    Layer(obj)
    self.failUnless(obj.hasExtension("App::GroupExtension"))

    if FreeCAD.GuiUp:
        LayerViewProvider(obj.ViewObject)
        self.failUnless(obj.ViewObject.hasExtension("Gui::ViewProviderGroupExtension"))
        self.failUnless(obj.ViewObject.hasExtension("Gui::ViewProviderGroupExtensionPython"))

    self.Doc.removeObject(obj.Name)
    del obj

  def testPropertyLink_Issue2902Part1(self):
    o1 = self.Doc.addObject("App::FeatureTest","test1")
    o2 = self.Doc.addObject("App::FeatureTest","test2")
    o3 = self.Doc.addObject("App::FeatureTest","test3")

    o1.Link=o2
    self.assertEqual(o1.Link, o2)
    o1.Link=o3
    self.assertEqual(o1.Link, o3)
    o2.Placement = FreeCAD.Placement()
    self.assertEqual(o1.Link, o3)

  def testNotification_Issue2902Part2(self):
    o = self.Doc.addObject("App::FeatureTest","test")

    plm = o.Placement
    o.Placement = FreeCAD.Placement()
    plm.Base.x = 5
    self.assertEqual(o.Placement.Base.x, 0)
    o.Placement.Base.x=5
    self.assertEqual(o.Placement.Base.x, 5)

  def testNotification_Issue2996(self):
    if not FreeCAD.GuiUp:
      return
    # works only if Gui is shown
    class ViewProvider:
      def __init__(self, vobj):
        vobj.Proxy=self

      def attach(self, vobj):
        self.ViewObject = vobj
        self.Object = vobj.Object

      def claimChildren(self):
        children = [self.Object.Link]
        return children

    obj=self.Doc.addObject("App::FeaturePython", "Sketch")
    obj.addProperty("App::PropertyLink","Link")
    ViewProvider(obj.ViewObject)

    ext=self.Doc.addObject("App::FeatureTest", "Extrude")
    ext.Link=obj

    sli=self.Doc.addObject("App::FeaturePython", "Slice")
    sli.addProperty("App::PropertyLink","Link").Link=ext
    ViewProvider(sli.ViewObject)

    com=self.Doc.addObject("App::FeaturePython", "CompoundFilter")
    com.addProperty("App::PropertyLink", "Link").Link=sli
    ViewProvider(com.ViewObject)

    ext.Label="test"

    self.assertEqual(ext.Link, obj)
    self.assertNotEqual(ext.Link, sli)

  def tearDown(self):
    #closing doc
    FreeCAD.closeDocument("CreateTest")

# class must be defined in global scope to allow it to be reloaded on document open
class SaveRestoreSpecialGroup():
    def __init__(self, obj):
        obj.addExtension("App::GroupExtensionPython", self)
        obj.Proxy = self

    def allowObject(self, obj):
        return False;

# class must be defined in global scope to allow it to be reloaded on document open
class SaveRestoreSpecialGroupViewProvider():
    def __init__(self, obj):
        obj.addExtension("Gui::ViewProviderGroupExtensionPython", self)
        obj.Proxy = self

    def testFunction(self):
        pass

class DocumentSaveRestoreCases(unittest.TestCase):
  def setUp(self):
    self.Doc = FreeCAD.newDocument("SaveRestoreTests")
    L1 = self.Doc.addObject("App::FeatureTest","Label_1")
    L2 = self.Doc.addObject("App::FeatureTest","Label_2")
    L3 = self.Doc.addObject("App::FeatureTest","Label_3")
    self.TempPath = tempfile.gettempdir()
    FreeCAD.Console.PrintLog( '  Using temp path: ' + self.TempPath + '\n')

  def testSaveAndRestore(self):
    # saving and restoring
    SaveName = self.TempPath + os.sep + "SaveRestoreTests.FCStd"
    self.failUnless(self.Doc.Label_1.TypeTransient == 4711)
    self.Doc.Label_1.TypeTransient = 4712
    # setup Linking
    self.Doc.Label_1.Link = self.Doc.Label_2
    self.Doc.Label_2.Link = self.Doc.Label_3
    self.Doc.Label_1.LinkSub = (self.Doc.Label_2,["Sub1","Sub2"])
    self.Doc.Label_2.LinkSub = (self.Doc.Label_3,["Sub3","Sub4"])
    # save the document
    self.Doc.saveAs(SaveName)
    FreeCAD.closeDocument("SaveRestoreTests")
    self.Doc = FreeCAD.open(SaveName)
    self.failUnless(self.Doc.Label_1.Integer == 4711)
    self.failUnless(self.Doc.Label_2.Integer == 4711)
    # test Linkage
    self.failUnless(self.Doc.Label_1.Link == self.Doc.Label_2)
    self.failUnless(self.Doc.Label_2.Link == self.Doc.Label_3)
    self.failUnless(self.Doc.Label_1.LinkSub == (self.Doc.Label_2,["Sub1","Sub2"]))
    self.failUnless(self.Doc.Label_2.LinkSub == (self.Doc.Label_3,["Sub3","Sub4"]))
    # do NOT save transient properties
    self.failUnless(self.Doc.Label_1.TypeTransient == 4711)
    self.failUnless(self.Doc == FreeCAD.getDocument(self.Doc.Name))

  def testRestore(self):
    Doc = FreeCAD.newDocument("RestoreTests")
    Doc.addObject("App::FeatureTest","Label_1")
    # saving and restoring
    FileName = self.TempPath + os.sep + "Test2.FCStd"
    Doc.saveAs(FileName)
    # restore must first clear the current content
    Doc.restore()
    self.failUnless(len(Doc.Objects) == 1)
    FreeCAD.closeDocument("RestoreTests")

  def testActiveDocument(self):
    # open 2nd doc
    Second = FreeCAD.newDocument("Active")
    FreeCAD.closeDocument("Active")
    try:
        # There might be no active document anymore
        # This also checks for dangling pointers
        Active = FreeCAD.activeDocument()
        # Second is still a valid object
        self.failUnless(Second != Active)
    except:
        # Okay, no document open
        self.failUnless(True)

  def testExtensionSaveRestore(self):
    # saving and restoring
    SaveName = self.TempPath + os.sep + "SaveRestoreExtensions.FCStd"
    Doc = FreeCAD.newDocument("SaveRestoreExtensions")
    #we try to create a normal python object and add an extension to it
    obj  = Doc.addObject("App::DocumentObject", "Obj")
    grp1 = Doc.addObject("App::DocumentObject", "Extension_1")
    grp2 = Doc.addObject("App::FeaturePython", "Extension_2")

    grp1.addExtension("App::GroupExtensionPython", None)
    SaveRestoreSpecialGroup(grp2)
    if FreeCAD.GuiUp:
        SaveRestoreSpecialGroupViewProvider(grp2.ViewObject)
    grp2.Group = [obj]

    Doc.saveAs(SaveName)
    FreeCAD.closeDocument("SaveRestoreExtensions")
    Doc = FreeCAD.open(SaveName)

    self.failUnless(Doc.Extension_1.hasExtension("App::GroupExtension"))
    self.failUnless(Doc.Extension_2.hasExtension("App::GroupExtension"))
    self.failUnless(Doc.Extension_1.ExtensionProxy is None)
    self.failUnless(Doc.Extension_2.ExtensionProxy is not None)
    self.failUnless(Doc.Extension_2.Group[0] is Doc.Obj)
    self.failUnless(hasattr(Doc.Extension_2.Proxy, 'allowObject'))
    self.failUnless(hasattr(Doc.Extension_2.ExtensionProxy, 'allowObject'))

    if FreeCAD.GuiUp:
      self.failUnless(Doc.Extension_2.ViewObject.hasExtension("Gui::ViewProviderGroupExtensionPython"))
      self.failUnless(hasattr(Doc.Extension_2.ViewObject.Proxy, 'testFunction'))
      self.failUnless(hasattr(Doc.Extension_2.ViewObject.ExtensionProxy, 'testFunction'))

    FreeCAD.closeDocument("SaveRestoreExtensions")

  def testPersistenceContentDump(self):
    #test smallest level... property
    self.Doc.Label_1.Vector = (1,2,3)
    dump = self.Doc.Label_1.dumpPropertyContent('Vector', Compression = 9)
    self.Doc.Label_2.restorePropertyContent('Vector', dump)
    self.assertEqual(self.Doc.Label_1.Vector, self.Doc.Label_2.Vector)

    #next higher: object
    self.Doc.Label_1.Distance = 12
    self.Doc.Label_1.String = 'test'
    dump = self.Doc.Label_1.dumpContent()
    self.Doc.Label_3.restoreContent(dump)
    self.assertEqual(self.Doc.Label_1.Distance, self.Doc.Label_3.Distance)
    self.assertEqual(self.Doc.Label_1.String, self.Doc.Label_3.String)

    #highest level: document
    dump = self.Doc.dumpContent(9)
    Doc = FreeCAD.newDocument("DumpTest")
    Doc.restoreContent(dump)
    self.assertEqual(len(self.Doc.Objects), len(Doc.Objects))
    self.assertEqual(self.Doc.Label_1.Distance, Doc.Label_1.Distance)
    self.assertEqual(self.Doc.Label_1.String, Doc.Label_1.String)
    self.assertEqual(self.Doc.Label_1.Vector, Doc.Label_1.Vector)
    FreeCAD.closeDocument("DumpTest")

  def tearDown(self):
    #closing doc
    FreeCAD.closeDocument("SaveRestoreTests")

class DocumentRecomputeCases(unittest.TestCase):
  def setUp(self):
    self.Doc = FreeCAD.newDocument("RecomputeTests")
    self.L1 = self.Doc.addObject("App::FeatureTest","Label_1")
    self.L2 = self.Doc.addObject("App::FeatureTest","Label_2")
    self.L3 = self.Doc.addObject("App::FeatureTest","Label_3")

  def testDescent(self):
    # testing the up and downstream stuff
    FreeCAD.Console.PrintLog("def testDescent(self):Testcase not implemented\n")
    self.L1.Link = self.L2
    self.L2.Link = self.L3

  def testRecompute(self):

    # sequence to test recompute behaviour
    #       L1---\    L7
    #      /  \   \    |
    #    L2   L3   \  L8
    #   /  \ /  \  /
    #  L4   L5   L6

    L1 = self.Doc.addObject("App::FeatureTest","Label_1")
    L2 = self.Doc.addObject("App::FeatureTest","Label_2")
    L3 = self.Doc.addObject("App::FeatureTest","Label_3")
    L4 = self.Doc.addObject("App::FeatureTest","Label_4")
    L5 = self.Doc.addObject("App::FeatureTest","Label_5")
    L6 = self.Doc.addObject("App::FeatureTest","Label_6")
    L7 = self.Doc.addObject("App::FeatureTest","Label_7")
    L8 = self.Doc.addObject("App::FeatureTest","Label_8")
    L1.LinkList = [L2,L3,L6]
    L2.Link = L4
    L2.LinkList = [L5]
    L3.LinkList = [L5,L6]
    L7.Link = L8 #make second root

    self.failUnless(L7 in self.Doc.RootObjects)
    self.failUnless(L1 in self.Doc.RootObjects)

    self.failUnless(len(self.Doc.Objects) == len(self.Doc.TopologicalSortedObjects))

    seqDic = {}
    i = 0
    for obj in self.Doc.TopologicalSortedObjects:
        seqDic[obj] = i
        print(obj)
        i += 1

    self.failUnless(seqDic[L2] > seqDic[L1])
    self.failUnless(seqDic[L3] > seqDic[L1])
    self.failUnless(seqDic[L5] > seqDic[L2])
    self.failUnless(seqDic[L5] > seqDic[L3])
    self.failUnless(seqDic[L5] > seqDic[L1])


    self.failUnless((0, 0, 0, 0, 0, 0)==(L1.ExecCount,L2.ExecCount,L3.ExecCount,L4.ExecCount,L5.ExecCount,L6.ExecCount))
    self.failUnless(self.Doc.recompute()==4)
    self.failUnless((1, 1, 1, 0, 0, 0)==(L1.ExecCount,L2.ExecCount,L3.ExecCount,L4.ExecCount,L5.ExecCount,L6.ExecCount))
    L5.enforceRecompute()
    self.failUnless((1, 1, 1, 0, 0, 0)==(L1.ExecCount,L2.ExecCount,L3.ExecCount,L4.ExecCount,L5.ExecCount,L6.ExecCount))
    self.failUnless(self.Doc.recompute()==4)
    self.failUnless((2, 2, 2, 0, 1, 0)==(L1.ExecCount,L2.ExecCount,L3.ExecCount,L4.ExecCount,L5.ExecCount,L6.ExecCount))
    L4.enforceRecompute()
    self.failUnless(self.Doc.recompute()==3)
    self.failUnless((3, 3, 2, 1, 1, 0)==(L1.ExecCount,L2.ExecCount,L3.ExecCount,L4.ExecCount,L5.ExecCount,L6.ExecCount))
    L5.enforceRecompute()
    self.failUnless(self.Doc.recompute()==4)
    self.failUnless((4, 4, 3, 1, 2, 0)==(L1.ExecCount,L2.ExecCount,L3.ExecCount,L4.ExecCount,L5.ExecCount,L6.ExecCount))
    L6.enforceRecompute()
    self.failUnless(self.Doc.recompute()==3)
    self.failUnless((5, 4, 4, 1, 2, 1)==(L1.ExecCount,L2.ExecCount,L3.ExecCount,L4.ExecCount,L5.ExecCount,L6.ExecCount))
    L2.enforceRecompute()
    self.failUnless(self.Doc.recompute()==2)
    self.failUnless((6, 5, 4, 1, 2, 1)==(L1.ExecCount,L2.ExecCount,L3.ExecCount,L4.ExecCount,L5.ExecCount,L6.ExecCount))
    L1.enforceRecompute()
    self.failUnless(self.Doc.recompute()==1)
    self.failUnless((7, 5, 4, 1, 2, 1)==(L1.ExecCount,L2.ExecCount,L3.ExecCount,L4.ExecCount,L5.ExecCount,L6.ExecCount))

    self.Doc.removeObject(L1.Name)
    self.Doc.removeObject(L2.Name)
    self.Doc.removeObject(L3.Name)
    self.Doc.removeObject(L4.Name)
    self.Doc.removeObject(L5.Name)
    self.Doc.removeObject(L6.Name)
    self.Doc.removeObject(L7.Name)
    self.Doc.removeObject(L8.Name)

  def tearDown(self):
    #closing doc
    FreeCAD.closeDocument("RecomputeTests")

class UndoRedoCases(unittest.TestCase):
  def setUp(self):
    self.Doc = FreeCAD.newDocument("UndoTest")
    self.Doc.UndoMode = 0
    self.Doc.addObject("App::FeatureTest","Base")
    self.Doc.addObject("App::FeatureTest","Del")
    self.Doc.getObject("Del").Integer  = 2

  def testUndoProperties(self):
    # switch on the Undo
    self.Doc.UndoMode = 1

    # first transaction
    self.Doc.openTransaction("Transaction1")
    self.Doc.addObject("App::FeatureTest","test1")
    self.Doc.getObject("test1").Integer  = 1
    self.Doc.getObject("test1").String   = "test1"
    self.Doc.getObject("test1").Float    = 1.0
    self.Doc.getObject("test1").Bool  = 1

    #self.Doc.getObject("test1").IntegerList  = 1
    #self.Doc.getObject("test1").FloatList  = 1.0

    #self.Doc.getObject("test1").Matrix  = (1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0)
    #self.Doc.getObject("test1").Vector  = (1.0,1.0,1.0)

    # second transaction
    self.Doc.openTransaction("Transaction2")
    self.Doc.getObject("test1").Integer  = 2
    self.Doc.getObject("test1").String   = "test2"
    self.Doc.getObject("test1").Float    = 2.0
    self.Doc.getObject("test1").Bool  = 0

    # switch on the Undo OFF
    self.Doc.UndoMode = 0

  def testUndoClear(self):
    # switch on the Undo
    self.Doc.UndoMode = 1
    self.assertEqual(self.Doc.UndoNames,[])
    self.assertEqual(self.Doc.UndoCount,0)
    self.assertEqual(self.Doc.RedoNames,[])
    self.assertEqual(self.Doc.RedoCount,0)

    self.Doc.openTransaction("Transaction1")
    # becomes the active object
    self.Doc.addObject("App::FeatureTest","test1")
    self.Doc.commitTransaction()
    # removes the active object
    self.Doc.undo()
    self.assertEqual(self.Doc.ActiveObject,None)
    # deletes the active object
    self.Doc.clearUndos()
    self.assertEqual(self.Doc.ActiveObject,None)

  def testUndo(self):
    # switch on the Undo
    self.Doc.UndoMode = 1
    self.assertEqual(self.Doc.UndoNames,[])
    self.assertEqual(self.Doc.UndoCount,0)
    self.assertEqual(self.Doc.RedoNames,[])
    self.assertEqual(self.Doc.RedoCount,0)

    # first transaction
    self.Doc.openTransaction("Transaction1")
    self.Doc.addObject("App::FeatureTest","test1")
    self.Doc.getObject("test1").Integer  = 1
    self.Doc.getObject("Del").Integer  = 1
    self.Doc.removeObject("Del")
    self.assertEqual(self.Doc.UndoNames,['Transaction1'])
    self.assertEqual(self.Doc.UndoCount,1)
    self.assertEqual(self.Doc.RedoNames,[])
    self.assertEqual(self.Doc.RedoCount,0)

    # second transaction
    self.Doc.openTransaction("Transaction2")
    # new behavior: no change, no transaction
    self.assertEqual(self.Doc.UndoNames,['Transaction1'])
    self.assertEqual(self.Doc.UndoCount,1)
    self.assertEqual(self.Doc.RedoNames,[])
    self.assertEqual(self.Doc.RedoCount,0)

    self.Doc.getObject("test1").Integer  = 2
    self.assertEqual(self.Doc.UndoNames,['Transaction2','Transaction1'])
    self.assertEqual(self.Doc.UndoCount,2)
    self.assertEqual(self.Doc.RedoNames,[])
    self.assertEqual(self.Doc.RedoCount,0)

    # abort second transaction
    self.Doc.abortTransaction()
    self.assertEqual(self.Doc.UndoNames,['Transaction1'])
    self.assertEqual(self.Doc.UndoCount,1)
    self.assertEqual(self.Doc.RedoNames,[])
    self.assertEqual(self.Doc.RedoCount,0)
    self.assertEqual(self.Doc.getObject("test1").Integer, 1)

    # again second transaction
    self.Doc.openTransaction("Transaction2")
    self.Doc.getObject("test1").Integer  = 2
    self.assertEqual(self.Doc.UndoNames,['Transaction2','Transaction1'])
    self.assertEqual(self.Doc.UndoCount,2)
    self.assertEqual(self.Doc.RedoNames,[])
    self.assertEqual(self.Doc.RedoCount,0)

    # third transaction
    self.Doc.openTransaction("Transaction3")
    self.Doc.getObject("test1").Integer  = 3
    self.assertEqual(self.Doc.UndoNames,['Transaction3','Transaction2','Transaction1'])
    self.assertEqual(self.Doc.UndoCount,3)
    self.assertEqual(self.Doc.RedoNames,[])
    self.assertEqual(self.Doc.RedoCount,0)

    # fourth transaction
    self.Doc.openTransaction("Transaction4")
    self.Doc.getObject("test1").Integer  = 4
    self.assertEqual(self.Doc.UndoNames,['Transaction4','Transaction3','Transaction2','Transaction1'])
    self.assertEqual(self.Doc.UndoCount,4)
    self.assertEqual(self.Doc.RedoNames,[])
    self.assertEqual(self.Doc.RedoCount,0)

    # undo the fourth transaction
    self.Doc.undo()
    self.assertEqual(self.Doc.getObject("test1").Integer, 3)
    self.assertEqual(self.Doc.UndoNames,['Transaction3','Transaction2','Transaction1'])
    self.assertEqual(self.Doc.UndoCount,3)
    self.assertEqual(self.Doc.RedoNames,['Transaction4'])
    self.assertEqual(self.Doc.RedoCount,1)

    # undo the third transaction
    self.Doc.undo()
    self.assertEqual(self.Doc.getObject("test1").Integer, 2)
    self.assertEqual(self.Doc.UndoNames,['Transaction2','Transaction1'])
    self.assertEqual(self.Doc.UndoCount,2)
    self.assertEqual(self.Doc.RedoNames,['Transaction3','Transaction4'])
    self.assertEqual(self.Doc.RedoCount,2)

    # undo the second transaction
    self.Doc.undo()
    self.assertEqual(self.Doc.getObject("test1").Integer, 1)
    self.assertEqual(self.Doc.UndoNames,['Transaction1'])
    self.assertEqual(self.Doc.UndoCount,1)
    self.assertEqual(self.Doc.RedoNames,['Transaction2','Transaction3','Transaction4'])
    self.assertEqual(self.Doc.RedoCount,3)

    # undo the first transaction
    self.Doc.undo()
    self.failUnless(self.Doc.getObject("test1") is None)
    self.failUnless(self.Doc.getObject("Del").Integer == 2)
    self.assertEqual(self.Doc.UndoNames,[])
    self.assertEqual(self.Doc.UndoCount,0)
    self.assertEqual(self.Doc.RedoNames,['Transaction1','Transaction2','Transaction3','Transaction4'])
    self.assertEqual(self.Doc.RedoCount,4)

    # redo the first transaction
    self.Doc.redo()
    self.assertEqual(self.Doc.getObject("test1").Integer, 1)
    self.assertEqual(self.Doc.UndoNames,['Transaction1'])
    self.assertEqual(self.Doc.UndoCount,1)
    self.assertEqual(self.Doc.RedoNames,['Transaction2','Transaction3','Transaction4'])
    self.assertEqual(self.Doc.RedoCount,3)

    # redo the second transaction
    self.Doc.redo()
    self.assertEqual(self.Doc.getObject("test1").Integer, 2)
    self.assertEqual(self.Doc.UndoNames,['Transaction2','Transaction1'])
    self.assertEqual(self.Doc.UndoCount,2)
    self.assertEqual(self.Doc.RedoNames,['Transaction3','Transaction4'])
    self.assertEqual(self.Doc.RedoCount,2)

    # undo the second transaction
    self.Doc.undo()
    self.assertEqual(self.Doc.getObject("test1").Integer, 1)
    self.assertEqual(self.Doc.UndoNames,['Transaction1'])
    self.assertEqual(self.Doc.UndoCount,1)
    self.assertEqual(self.Doc.RedoNames,['Transaction2','Transaction3','Transaction4'])
    self.assertEqual(self.Doc.RedoCount,3)

    # new transaction eight
    self.Doc.openTransaction("Transaction8")
    self.Doc.getObject("test1").Integer  = 8
    self.assertEqual(self.Doc.UndoNames,['Transaction8','Transaction1'])
    self.assertEqual(self.Doc.UndoCount,2)
    self.assertEqual(self.Doc.RedoNames,[])
    self.assertEqual(self.Doc.RedoCount,0)
    self.Doc.abortTransaction()
    self.assertEqual(self.Doc.UndoNames,['Transaction1'])
    self.assertEqual(self.Doc.UndoCount,1)
    self.assertEqual(self.Doc.RedoNames,[])
    self.assertEqual(self.Doc.RedoCount,0)

    # again new transaction eight
    self.Doc.openTransaction("Transaction8")
    self.Doc.getObject("test1").Integer  = 8
    self.assertEqual(self.Doc.UndoNames,['Transaction8','Transaction1'])
    self.assertEqual(self.Doc.UndoCount,2)
    self.assertEqual(self.Doc.RedoNames,[])
    self.assertEqual(self.Doc.RedoCount,0)

    # again new transaction nine
    self.Doc.openTransaction("Transaction9")
    self.Doc.getObject("test1").Integer  = 9
    self.assertEqual(self.Doc.UndoNames,['Transaction9','Transaction8','Transaction1'])
    self.assertEqual(self.Doc.UndoCount,3)
    self.assertEqual(self.Doc.RedoNames,[])
    self.assertEqual(self.Doc.RedoCount,0)
    self.Doc.commitTransaction()
    self.assertEqual(self.Doc.UndoNames,['Transaction9','Transaction8','Transaction1'])
    self.assertEqual(self.Doc.UndoCount,3)
    self.assertEqual(self.Doc.RedoNames,[])
    self.assertEqual(self.Doc.RedoCount,0)
    self.assertEqual(self.Doc.getObject("test1").Integer, 9)

    # undo the ninth transaction
    self.Doc.undo()
    self.assertEqual(self.Doc.getObject("test1").Integer, 8)
    self.assertEqual(self.Doc.UndoNames,['Transaction8','Transaction1'])
    self.assertEqual(self.Doc.UndoCount,2)
    self.assertEqual(self.Doc.RedoNames,['Transaction9'])
    self.assertEqual(self.Doc.RedoCount,1)

    # switch on the Undo OFF
    self.Doc.UndoMode = 0
    self.assertEqual(self.Doc.UndoNames,[])
    self.assertEqual(self.Doc.UndoCount,0)
    self.assertEqual(self.Doc.RedoNames,[])
    self.assertEqual(self.Doc.RedoCount,0)

  def testUndoInList(self):

    self.Doc.UndoMode = 1

    self.Doc.openTransaction("Box")
    self.Box = self.Doc.addObject('App::FeatureTest')
    self.Doc.commitTransaction()

    self.Doc.openTransaction("Cylinder")
    self.Cylinder = self.Doc.addObject('App::FeatureTest')
    self.Doc.commitTransaction()

    self.Doc.openTransaction("Fuse")
    self.Fuse1 = self.Doc.addObject('App::FeatureTest', 'Fuse')
    self.Fuse1.LinkList = [self.Box, self.Cylinder]
    self.Doc.commitTransaction()

    self.Doc.undo()
    self.failUnless(len(self.Box.InList) == 0)
    self.failUnless(len(self.Cylinder.InList) == 0)

    self.Doc.redo()
    self.failUnless(len(self.Box.InList) == 1)
    self.failUnless(self.Box.InList[0] == self.Doc.Fuse)
    self.failUnless(len(self.Cylinder.InList) == 1)
    self.failUnless(self.Cylinder.InList[0] == self.Doc.Fuse)

  def testUndoIssue0003150Part1(self):

    self.Doc.UndoMode = 1

    self.Doc.openTransaction("Box")
    self.Box = self.Doc.addObject('App::FeatureTest')
    self.Doc.commitTransaction()

    self.Doc.openTransaction("Cylinder")
    self.Cylinder = self.Doc.addObject('App::FeatureTest')
    self.Doc.commitTransaction()

    self.Doc.openTransaction("Fuse")
    self.Fuse1 = self.Doc.addObject('App::FeatureTest')
    self.Fuse1.LinkList = [self.Box, self.Cylinder]
    self.Doc.commitTransaction()
    self.Doc.recompute()

    self.Doc.openTransaction("Sphere")
    self.Sphere = self.Doc.addObject('App::FeatureTest')
    self.Doc.commitTransaction()

    self.Doc.openTransaction("Fuse")
    self.Fuse2 = self.Doc.addObject('App::FeatureTest')
    self.Fuse2.LinkList = [self.Fuse1, self.Sphere]
    self.Doc.commitTransaction()
    self.Doc.recompute()

    self.Doc.openTransaction("Part")
    self.Part = self.Doc.addObject('App::Part')
    self.Doc.commitTransaction()

    self.Doc.openTransaction("Drag")
    self.Part.addObject(self.Fuse2)
    self.Doc.commitTransaction()

    #3 undos show the problem of failing recompute
    self.Doc.undo()
    self.Doc.undo()
    self.Doc.undo()
    self.failUnless(self.Doc.recompute() >= 0)

  def tearDown(self):
    # closing doc
    FreeCAD.closeDocument("UndoTest")

class DocumentGroupCases(unittest.TestCase):

  def setUp(self):
    self.Doc = FreeCAD.newDocument("GroupTests")

  def testGroup(self):
    # Add an object to the group
    L2 = self.Doc.addObject("App::FeatureTest","Label_2")
    G1 = self.Doc.addObject("App::DocumentObjectGroup","Group")
    G1.addObject(L2)
    self.failUnless(G1.hasObject(L2))

    # Adding the group to itself must fail
    try:
      G1.addObject(G1)
    except:
      FreeCAD.Console.PrintLog("Cannot add group to itself, OK\n")
    else:
      self.fail("Adding the group to itself must not be possible")

    self.Doc.UndoMode = 1

    # Remove object from group
    self.Doc.openTransaction("Remove")
    self.Doc.removeObject("Label_2")
    self.Doc.commitTransaction()
    self.failUnless(G1.getObject("Label_2") is None)
    self.Doc.undo()
    self.failUnless(G1.getObject("Label_2") != None)

    # Remove first group and then the object
    self.Doc.openTransaction("Remove")
    self.Doc.removeObject("Group")
    self.Doc.removeObject("Label_2")
    self.Doc.commitTransaction()
    self.Doc.undo()
    self.failUnless(G1.getObject("Label_2") != None)

    # Remove first object and then the group in two transactions
    self.Doc.openTransaction("Remove")
    self.Doc.removeObject("Label_2")
    self.Doc.commitTransaction()
    self.failUnless(G1.getObject("Label_2") is None)
    self.Doc.openTransaction("Remove")
    self.Doc.removeObject("Group")
    self.Doc.commitTransaction()
    self.Doc.undo()
    self.Doc.undo()
    self.failUnless(G1.getObject("Label_2") != None)

    # Remove first object and then the group in one transaction
    self.Doc.openTransaction("Remove")
    self.Doc.removeObject("Label_2")
    self.failUnless(G1.getObject("Label_2") is None)
    self.Doc.removeObject("Group")
    self.Doc.commitTransaction()
    self.Doc.undo()
    # FIXME: See bug #1820554
    self.failUnless(G1.getObject("Label_2") != None)

    # Add a second object to the group
    L3 = self.Doc.addObject("App::FeatureTest","Label_3")
    G1.addObject(L3)
    self.Doc.openTransaction("Remove")
    self.Doc.removeObject("Label_2")
    self.failUnless(G1.getObject("Label_2") is None)
    self.Doc.removeObject("Label_3")
    self.failUnless(G1.getObject("Label_3") is None)
    self.Doc.removeObject("Group")
    self.Doc.commitTransaction()
    self.Doc.undo()
    self.failUnless(G1.getObject("Label_3") != None)
    self.failUnless(G1.getObject("Label_2") != None)

    self.Doc.UndoMode = 0

    # Cleanup
    self.Doc.removeObject("Group")
    self.Doc.removeObject("Label_2")
    self.Doc.removeObject("Label_3")

  def testGroupAndGeoFeatureGroup(self):

    # an object can only be in one group at once, that must be enforced
    obj1 = self.Doc.addObject("App::FeatureTest","obj1")
    grp1 = self.Doc.addObject("App::DocumentObjectGroup","Group1")
    grp2 = self.Doc.addObject("App::DocumentObjectGroup","Group2")
    grp1.addObject(obj1)
    self.failUnless(obj1.getParentGroup()==grp1)
    self.failUnless(obj1.getParentGeoFeatureGroup() is None)
    self.failUnless(grp1.hasObject(obj1))
    grp2.addObject(obj1)
    self.failUnless(grp1.hasObject(obj1)==False)
    self.failUnless(grp2.hasObject(obj1))

    # an object is allowed to be in a group and a geofeaturegroup
    prt1 = self.Doc.addObject("App::Part","Part1")
    prt2 = self.Doc.addObject("App::Part","Part2")

    prt1.addObject(grp2)
    self.failUnless(grp2.getParentGeoFeatureGroup() == prt1)
    self.failUnless(grp2.getParentGroup() is None)
    self.failUnless(grp2.hasObject(obj1))
    self.failUnless(prt1.hasObject(grp2))
    self.failUnless(prt1.hasObject(obj1))

    #it is not allowed to be in 2 geofeaturegroups
    prt2.addObject(grp2)
    self.failUnless(grp2.hasObject(obj1))
    self.failUnless(prt1.hasObject(grp2)==False)
    self.failUnless(prt1.hasObject(obj1)==False)
    self.failUnless(prt2.hasObject(grp2))
    self.failUnless(prt2.hasObject(obj1))
    try:
        grp = prt1.Group
        grp.append(obj1)
        prt1.Group = grp
    except:
        grp.remove(obj1)
        self.failUnless(prt1.Group == grp)
    else:
        self.fail("No exception thrown when object is in multiple Groups")

    #it is not allowed to be in 2 Groups
    prt2.addObject(grp1)
    grp = grp1.Group
    grp.append(obj1)
    try:
        grp1.Group = grp
    except:
        pass
    else:
        self.fail("No exception thrown when object is in multiple Groups")

    #cross linking between GeoFeatureGroups is not allowed
    self.Doc.recompute()
    box = self.Doc.addObject("App::FeatureTest","Box")
    cyl = self.Doc.addObject("App::FeatureTest","Cylinder")
    fus = self.Doc.addObject("App::FeatureTest","Fusion")
    fus.LinkList = [cyl, box]
    self.Doc.recompute()
    self.failUnless(fus.State[0] == 'Up-to-date')
    fus.LinkList = [] #remove all links as addObject would otherwise transfer all linked objects
    prt1.addObject(cyl)
    fus.LinkList = [cyl, box]
    self.Doc.recompute()
    #self.failUnless(fus.State[0] == 'Invalid')
    fus.LinkList = []
    prt1.addObject(box)
    fus.LinkList = [cyl, box]
    self.Doc.recompute()
    #self.failUnless(fus.State[0] == 'Invalid')
    fus.LinkList = []
    prt1.addObject(fus)
    fus.LinkList = [cyl, box]
    self.Doc.recompute()
    self.failUnless(fus.State[0] == 'Up-to-date')
    prt2.addObject(box) #this time addObject should move all dependencies to the new part
    self.Doc.recompute()
    self.failUnless(fus.State[0] == 'Up-to-date')

    #grouping must be resilient against cyclic links and not crash: #issue 0002567
    prt1.addObject(prt2)
    grp = prt2.Group
    grp.append(prt1)
    prt2.Group = grp
    self.Doc.recompute()
    prt2.Group = []
    try:
        prt2.Group = [prt2]
    except:
        pass
    else:
        self.fail("Exception is expected")

    self.Doc.recompute()

  def testIssue0003150Part2(self):
    self.box = self.Doc.addObject("App::FeatureTest")
    self.cyl = self.Doc.addObject("App::FeatureTest")
    self.sph = self.Doc.addObject("App::FeatureTest")

    self.fus1 = self.Doc.addObject("App::FeatureTest")
    self.fus2 = self.Doc.addObject("App::FeatureTest")

    self.fus1.LinkList = [self.box, self.cyl];
    self.fus2.LinkList = [self.sph, self.cyl];

    self.prt = self.Doc.addObject("App::Part")
    self.prt.addObject(self.fus1)
    self.failUnless(len(self.prt.Group)==5)
    self.failUnless(self.fus2.getParentGeoFeatureGroup() == self.prt)
    self.failUnless(self.prt.hasObject(self.sph))

    self.prt.removeObject(self.fus1)
    self.failUnless(len(self.prt.Group)==0)

  def tearDown(self):
    # closing doc
    FreeCAD.closeDocument("GroupTests")

class DocumentPlatformCases(unittest.TestCase):
  def setUp(self):
    self.Doc = FreeCAD.newDocument("PlatformTests")
    self.Doc.addObject("App::FeatureTest", "Test")
    self.TempPath = tempfile.gettempdir()
    self.DocName = self.TempPath + os.sep + "PlatformTests.FCStd"

  def testFloatList(self):
    self.Doc.Test.FloatList = [-0.05, 2.5, 5.2]

    # saving and restoring
    self.Doc.saveAs(self.DocName)
    FreeCAD.closeDocument("PlatformTests")
    self.Doc = FreeCAD.open(self.DocName)

    self.failUnless(abs(self.Doc.Test.FloatList[0] + .05) < 0.01)
    self.failUnless(abs(self.Doc.Test.FloatList[1] - 2.5) < 0.01)
    self.failUnless(abs(self.Doc.Test.FloatList[2] - 5.2) < 0.01)

  def testColorList(self):
    self.Doc.Test.ColourList = [(1.0,0.5,0.0),(0.0,0.5,1.0)]

    # saving and restoring
    self.Doc.saveAs(self.DocName)
    FreeCAD.closeDocument("PlatformTests")
    self.Doc = FreeCAD.open(self.DocName)

    self.failUnless(abs(self.Doc.Test.ColourList[0][0] - 1.0) < 0.01)
    self.failUnless(abs(self.Doc.Test.ColourList[0][1] - 0.5) < 0.01)
    self.failUnless(abs(self.Doc.Test.ColourList[0][2] - 0.0) < 0.01)
    self.failUnless(abs(self.Doc.Test.ColourList[0][3] - 0.0) < 0.01)
    self.failUnless(abs(self.Doc.Test.ColourList[1][0] - 0.0) < 0.01)
    self.failUnless(abs(self.Doc.Test.ColourList[1][1] - 0.5) < 0.01)
    self.failUnless(abs(self.Doc.Test.ColourList[1][2] - 1.0) < 0.01)
    self.failUnless(abs(self.Doc.Test.ColourList[1][3] - 0.0) < 0.01)

  def testVectorList(self):
    self.Doc.Test.VectorList = [(-0.05, 2.5, 5.2),(-0.05, 2.5, 5.2)]

    # saving and restoring
    self.Doc.saveAs(self.DocName)
    FreeCAD.closeDocument("PlatformTests")
    self.Doc = FreeCAD.open(self.DocName)

    self.failUnless(len(self.Doc.Test.VectorList) == 2)

  def testPoints(self):
    try:
      self.Doc.addObject("Points::Feature", "Points")

      # saving and restoring
      self.Doc.saveAs(self.DocName)
      FreeCAD.closeDocument("PlatformTests")
      self.Doc = FreeCAD.open(self.DocName)

      self.failUnless(self.Doc.Points.Points.count() == 0)
    except:
      pass

  def tearDown(self):
    #closing doc
    FreeCAD.closeDocument("PlatformTests")


class DocumentBacklinks(unittest.TestCase):
  def setUp(self):
    self.Doc = FreeCAD.newDocument("BackLinks")

  def testIssue0003323(self):
    self.Doc.UndoMode=1
    self.Doc.openTransaction("Create object")
    obj1=self.Doc.addObject("App::FeatureTest","Test1")
    obj2=self.Doc.addObject("App::FeatureTest","Test2")
    obj2.Link=obj1
    self.Doc.commitTransaction()
    self.Doc.undo()
    self.Doc.openTransaction("Create object")

  def tearDown(self):
    # closing doc
    FreeCAD.closeDocument("BackLinks")


class DocumentFileIncludeCases(unittest.TestCase):
  def setUp(self):
    self.Doc = FreeCAD.newDocument("FileIncludeTests")
    # testing with undo
    self.Doc.UndoMode = 1


  def testApplyFiles(self):
    self.Doc.openTransaction("Transaction0")
    self.L1 = self.Doc.addObject("App::DocumentObjectFileIncluded","FileObject1")
    self.failUnless(self.L1.File =="")
    self.Filename = self.L1.File

    self.Doc.openTransaction("Transaction1")
    self.TempPath = tempfile.gettempdir()
    # creating a file in the Transient directory of the document
    file = open(self.Doc.getTempFileName("test"),"w")
    file.write("test No1")
    file.close()
    # applying the file
    self.L1.File = (file.name,"Test.txt")
    self.failUnless(self.L1.File.split("/")[-1] == "Test.txt")
    # read again
    file = open(self.L1.File,"r")
    self.failUnless(file.read()=="test No1")
    file.close()
    file = open(self.TempPath+"/testNest.txt","w")
    file.write("test No2")
    file.close()
    # applying the file
    self.Doc.openTransaction("Transaction2")
    self.L1.File = file.name
    self.failUnless(self.L1.File.split("/")[-1] == "Test.txt")
    # read again
    file = open(self.L1.File,"r")
    self.failUnless(file.read()=="test No2")
    file.close()
    self.Doc.undo()
    self.failUnless(self.L1.File.split("/")[-1] == "Test.txt")
    # read again
    file = open(self.L1.File,"r")
    self.failUnless(file.read()=="test No1")
    file.close()
    self.Doc.undo()
    # read again
    self.failUnless(self.L1.File == "")
    self.Doc.redo()
    self.failUnless(self.L1.File.split("/")[-1] == "Test.txt")
    # read again
    file = open(self.L1.File,"r")
    self.failUnless(file.read()=="test No1")
    file.close()
    self.Doc.redo()
    self.failUnless(self.L1.File.split("/")[-1] == "Test.txt")
    # read again
    file = open(self.L1.File,"r")
    self.failUnless(file.read()=="test No2")
    file.close()
    # Save restore test
    FileName = self.TempPath+"/FileIncludeTests.fcstd"
    self.Doc.saveAs(FileName)
    FreeCAD.closeDocument("FileIncludeTests")
    self.Doc = FreeCAD.open(self.TempPath+"/FileIncludeTests.fcstd")
    # check if the file is still there
    self.L1 = self.Doc.getObject("FileObject1")
    file = open(self.L1.File,"r")
    res = file.read()
    FreeCAD.Console.PrintLog( res +"\n")
    self.failUnless(res=="test No2")
    self.failUnless(self.L1.File.split("/")[-1] == "Test.txt")
    file.close()

    # test for bug #94 (File overlap in PropertyFileIncluded)
    L2 = self.Doc.addObject("App::DocumentObjectFileIncluded","FileObject2")
    L3 = self.Doc.addObject("App::DocumentObjectFileIncluded","FileObject3")

    # creating two files in the Transient directory of the document
    file1 = open(self.Doc.getTempFileName("test"),"w")
    file1.write("test No1")
    file1.close()
    file2 = open(self.Doc.getTempFileName("test"),"w")
    file2.write("test No2")
    file2.close()

    # applying the file with the same base name
    L2.File = (file1.name,"Test.txt")
    L3.File = (file2.name,"Test.txt")

    file = open(L2.File,"r")
    self.failUnless(file.read()=="test No1")
    file.close()
    file = open(L3.File,"r")
    self.failUnless(file.read()=="test No2")
    file.close()

    # create a second document, copy a file and close the document
    # the test is about to put the file to the correct transient dir
    doc2 = FreeCAD.newDocument("Doc2")
    L4 = doc2.addObject("App::DocumentObjectFileIncluded","FileObject")
    L5 = doc2.addObject("App::DocumentObjectFileIncluded","FileObject")
    L6 = doc2.addObject("App::DocumentObjectFileIncluded","FileObject")
    L4.File = (L3.File,"Test.txt")
    L5.File = L3.File
    L6.File = L3.File
    FreeCAD.closeDocument("FileIncludeTests")
    self.Doc = FreeCAD.open(self.TempPath+"/FileIncludeTests.fcstd")
    self.failUnless(os.path.exists(L4.File))
    self.failUnless(os.path.exists(L5.File))
    self.failUnless(os.path.exists(L6.File))
    self.failUnless(L5.File != L6.File)
    # copy file from L5 which is in the same directory
    L7 = doc2.addObject("App::DocumentObjectFileIncluded","FileObject3")
    L7.File = (L5.File,"Copy.txt")
    self.failUnless(os.path.exists(L5.File))
    FreeCAD.closeDocument("Doc2")


  def tearDown(self):
    #closing doc
    FreeCAD.closeDocument("FileIncludeTests")


class DocumentPropertyCases(unittest.TestCase):
  def setUp(self):
    self.Doc = FreeCAD.newDocument("PropertyTests")
    self.Obj = self.Doc.addObject("App::FeaturePython","Test")

  def testDescent(self):
    # testing the up and downstream stuff
    props=self.Obj.supportedProperties()
    for i in props:
        self.Obj.addProperty(i,i)
    tempPath = tempfile.gettempdir()
    tempFile = tempPath + os.sep + "PropertyTests.FCStd"
    self.Doc.saveAs(tempFile)
    FreeCAD.closeDocument("PropertyTests")
    self.Doc = FreeCAD.open(tempFile)

  def testRemoveProperty(self):
    prop = 'Something'
    self.Obj.addProperty('App::PropertyFloat', prop)
    self.Obj.Something = 0.01
    self.Doc.recompute()
    self.Doc.openTransaction('modify and remove property')
    self.Obj.Something = 0.00
    self.Obj.removeProperty(prop)
    self.Obj.recompute()
    self.Doc.abortTransaction()

  def testRemovePropertyExpression(self):
    p1 = self.Doc.addObject("App::FeaturePython", "params1")
    p2 = self.Doc.addObject("App::FeaturePython", "params2")
    p1.addProperty("App::PropertyFloat", "a")
    p1.a = 42
    p2.addProperty("App::PropertyFloat", "b")
    p2.setExpression('b', u'params1.a')
    self.Doc.recompute()
    p2.removeProperty("b")
    p1.touch()
    self.Doc.recompute()
    self.assertTrue(not p2 in p1.InList)

  def testRemovePropertyOnChange(self):
    class Feature:
      def __init__(self, fp):
        fp.Proxy = self
        fp.addProperty("App::PropertyString","Test")
      def onBeforeChange(self, fp, prop):
        if prop == "Test":
          fp.removeProperty("Test")
      def onChanged(self, fp, prop):
        getattr(fp, prop)

    obj = self.Doc.addObject("App::FeaturePython")
    fea = Feature(obj)
    obj.Test = "test"

  def tearDown(self):
    #closing doc
    FreeCAD.closeDocument("PropertyTests")


class DocumentExpressionCases(unittest.TestCase):
  def setUp(self):
    self.Doc = FreeCAD.newDocument()
    self.Obj1 = self.Doc.addObject("App::FeatureTest","Test")
    self.Obj2 = self.Doc.addObject("App::FeatureTest","Test")

  def assertAlmostEqual (self, v1, v2) :
    if (math.fabs(v2-v1) > 1E-12) :
      self.assertEqual(v1,v2)


  def testExpression(self):
    # set the object twice to test that the backlinks are removed when overwriting the expression
    self.Obj2.setExpression('Placement.Rotation.Angle', u'%s.Placement.Rotation.Angle' % self.Obj1.Name)
    self.Obj2.setExpression('Placement.Rotation.Angle', u'%s.Placement.Rotation.Angle' % self.Obj1.Name)
    self.Obj1.Placement = FreeCAD.Placement(FreeCAD.Vector(0,0,0),FreeCAD.Rotation(FreeCAD.Vector(0,0,1),10))
    self.Doc.recompute()
    self.assertAlmostEqual(self.Obj1.Placement.Rotation.Angle, self.Obj2.Placement.Rotation.Angle)

    # clear the expression
    self.Obj2.setExpression('Placement.Rotation.Angle', None)
    self.assertAlmostEqual(self.Obj1.Placement.Rotation.Angle, self.Obj2.Placement.Rotation.Angle)
    self.Doc.recompute()
    self.assertAlmostEqual(self.Obj1.Placement.Rotation.Angle, self.Obj2.Placement.Rotation.Angle)
    # touch the objects to perform a recompute
    self.Obj1.Placement = self.Obj1.Placement
    self.Obj2.Placement = self.Obj2.Placement
    # must not raise a topological error
    self.assertEqual(self.Doc.recompute(), 2)

  def tearDown(self):
    #closing doc
    FreeCAD.closeDocument(self.Doc.Name)


class DocumentObserverCases(unittest.TestCase):

  class Observer():
    
    def __init__(self):
      self.signal = []
      self.parameter = []
      self.parameter2 = []
    
    def slotCreatedDocument(self, doc):
      self.signal.append('DocCreated');
      self.parameter.append(doc);
      
    def slotDeletedDocument(self, doc):
      self.signal.append('DocDeleted');
      self.parameter.append(doc);
      
    def slotRelabelDocument(self, doc):
      self.signal.append('DocRelabled');
      self.parameter.append(doc);
      
    def slotActivateDocument(self, doc):
      self.signal.append('DocActivated');
      self.parameter.append(doc);
      
    def slotRecomputedDocument(self, doc):
      self.signal.append('DocRecomputed');
      self.parameter.append(doc);

    def slotUndoDocument(self, doc):
      self.signal.append('DocUndo');
      self.parameter.append(doc);
      
    def slotRedoDocument(self, doc):
      self.signal.append('DocRedo');
      self.parameter.append(doc);

    def slotOpenTransaction(self, doc, name):
      self.signal.append('DocOpenTransaction');
      self.parameter.append(doc);
      self.parameter2.append(name);
      
    def slotCommitTransaction(self, doc):
      self.signal.append('DocCommitTransaction');
      self.parameter.append(doc);
      
    def slotAbortTransaction(self, doc):
      self.signal.append('DocAbortTransaction');
      self.parameter.append(doc);
     
    def slotBeforeChangeDocument(self, doc, prop):
        self.signal.append('DocBeforeChange')
        self.parameter.append(doc)
        self.parameter2.append(prop)
        
    def slotChangedDocument(self, doc, prop):
        self.signal.append('DocChanged')
        self.parameter.append(doc)
        self.parameter2.append(prop)
      
    def slotCreatedObject(self, obj):
      self.signal.append('ObjCreated');
      self.parameter.append(obj);

    def slotDeletedObject(self, obj):
      self.signal.append('ObjDeleted');
      self.parameter.append(obj)

    def slotChangedObject(self, obj, prop):
      self.signal.append('ObjChanged');
      self.parameter.append(obj)
      self.parameter2.append(prop)
      
    def slotBeforeChangeObject(self, obj, prop):
      self.signal.append('ObjBeforeChange');
      self.parameter.append(obj)
      self.parameter2.append(prop)

    def slotRecomputedObject(self, obj):
      self.signal.append('ObjRecomputed');
      self.parameter.append(obj)
      
    def slotAppendDynamicProperty(self, obj, prop):
      self.signal.append('ObjAddDynProp');
      self.parameter.append(obj)
      self.parameter2.append(prop)
    
    def slotRemoveDynamicProperty(self, obj, prop):
      self.signal.append('ObjRemoveDynProp');
      self.parameter.append(obj)
      self.parameter2.append(prop)
    
    def slotChangePropertyEditor(self, obj, prop):
      self.signal.append('ObjChangePropEdit');
      self.parameter.append(obj)
      self.parameter2.append(prop)

    def slotStartSaveDocument(self, obj, name):
      self.signal.append('DocStartSave')
      self.parameter.append(obj)
      self.parameter2.append(name)

    def slotFinishSaveDocument(self, obj, name):
      self.signal.append('DocFinishSave')
      self.parameter.append(obj)
      self.parameter2.append(name)
      
    def slotBeforeAddingDynamicExtension(self, obj, extension):
      self.signal.append('ObjBeforeDynExt')
      self.parameter.append(obj)
      self.parameter2.append(extension)
      
    def slotAddedDynamicExtension(self, obj, extension):
      self.signal.append('ObjDynExt')
      self.parameter.append(obj)
      self.parameter2.append(extension)

  class GuiObserver():
    
    def __init__(self):
      self.signal = []
      self.parameter = []
      self.parameter2 = []
    
    def slotCreatedDocument(self, doc):
      self.signal.append('DocCreated');
      self.parameter.append(doc);
      
    def slotDeletedDocument(self, doc):
      self.signal.append('DocDeleted');
      self.parameter.append(doc);
      
    def slotRelabelDocument(self, doc):
      self.signal.append('DocRelabled');
      self.parameter.append(doc);
      
    def slotRenameDocument(self, doc):
      self.signal.append('DocRenamed');
      self.parameter.append(doc);
      
    def slotActivateDocument(self, doc):
      self.signal.append('DocActivated');
      self.parameter.append(doc);
      
    def slotCreatedObject(self, obj):
      self.signal.append('ObjCreated');
      self.parameter.append(obj);

    def slotDeletedObject(self, obj):
      self.signal.append('ObjDeleted');
      self.parameter.append(obj)

    def slotChangedObject(self, obj, prop):
      self.signal.append('ObjChanged');
      self.parameter.append(obj)
      self.parameter2.append(prop)
      
    def slotInEdit(self, obj):
      self.signal.append('ObjInEdit');
      self.parameter.append(obj)  
    
    def slotResetEdit(self, obj):
      self.signal.append('ObjResetEdit');
      self.parameter.append(obj) 
      
  def setUp(self):
    self.Obs = self.Observer();
    FreeCAD.addDocumentObserver(self.Obs);

  def testRemoveObserver(self):
    FreeCAD.removeDocumentObserver(self.Obs)
    self.Obs.signal = []
    self.Obs.parameter = []
    self.Obs.parameter2 = []
    self.Doc1 = FreeCAD.newDocument("Observer")
    FreeCAD.closeDocument(self.Doc1.Name)
    self.assertEqual(len(self.Obs.signal), 0)
    self.assertEqual(len(self.Obs.parameter2), 0)
    self.assertEqual(len(self.Obs.signal), 0)
    FreeCAD.addDocumentObserver(self.Obs);

  def testSave(self):
    TempPath = tempfile.gettempdir()
    SaveName = TempPath + os.sep + "SaveRestoreTests.FCStd"
    self.Doc1 = FreeCAD.newDocument("Observer1");
    self.Doc1.saveAs(SaveName)
    self.assertEqual(self.Obs.signal.pop(), 'DocFinishSave')
    self.assertEqual(self.Obs.parameter2.pop(), self.Doc1.FileName)
    self.assertEqual(self.Obs.signal.pop(), 'DocStartSave')
    self.assertEqual(self.Obs.parameter2.pop(), self.Doc1.FileName)
    FreeCAD.closeDocument(self.Doc1.Name)

  def testDocument(self):
    # in case another document already exists then the tests cannot
    # be done reliably
    if FreeCAD.GuiUp and FreeCAD.activeDocument():
      return

    # testing document level signals
    self.Doc1 = FreeCAD.newDocument("Observer1");  
    if FreeCAD.GuiUp:
      self.assertEqual(self.Obs.signal.pop(0), 'DocActivated')
      self.assertTrue(self.Obs.parameter.pop(0) is self.Doc1)
    self.assertEqual(self.Obs.signal.pop(0), 'DocCreated')
    self.assertTrue(self.Obs.parameter.pop(0) is self.Doc1)
    self.assertEqual(self.Obs.signal.pop(0), 'DocBeforeChange')
    self.assertTrue(self.Obs.parameter.pop(0) is self.Doc1)
    self.assertEqual(self.Obs.parameter2.pop(0), 'Label')
    self.assertEqual(self.Obs.signal.pop(0), 'DocChanged')
    self.assertTrue(self.Obs.parameter.pop(0) is self.Doc1)
    self.assertEqual(self.Obs.parameter2.pop(0), 'Label')
    self.assertEqual(self.Obs.signal.pop(0), 'DocRelabled')
    self.assertTrue(self.Obs.parameter.pop(0) is self.Doc1)
    self.assertTrue(not self.Obs.signal and not self.Obs.parameter and not self.Obs.parameter2)
    
    self.Doc2 = FreeCAD.newDocument("Observer2");
    if FreeCAD.GuiUp:
      self.assertEqual(self.Obs.signal.pop(0), 'DocActivated')
      self.assertTrue(self.Obs.parameter.pop(0) is self.Doc2)
    self.assertEqual(self.Obs.signal.pop(0), 'DocCreated')
    self.assertTrue(self.Obs.parameter.pop(0) is self.Doc2)
    self.assertEqual(self.Obs.signal.pop(0), 'DocBeforeChange')
    self.assertTrue(self.Obs.parameter.pop(0) is self.Doc2)
    self.assertEqual(self.Obs.parameter2.pop(0), 'Label')
    self.assertEqual(self.Obs.signal.pop(0), 'DocChanged')
    self.assertTrue(self.Obs.parameter.pop(0) is self.Doc2)
    self.assertEqual(self.Obs.parameter2.pop(0), 'Label')
    self.assertEqual(self.Obs.signal.pop(0), 'DocRelabled')
    self.assertTrue(self.Obs.parameter.pop(0) is self.Doc2)
    self.assertTrue(not self.Obs.signal and not self.Obs.parameter and not self.Obs.parameter2)
    
    FreeCAD.setActiveDocument('Observer1')
    self.assertEqual(self.Obs.signal.pop(), 'DocActivated')
    self.assertTrue(self.Obs.parameter.pop() is self.Doc1)
    self.assertTrue(not self.Obs.signal and not self.Obs.parameter and not self.Obs.parameter2)

    #undo/redo is not enabled in cmd line mode by default
    self.Doc2.UndoMode = 1
    
    # Must set Doc2 as active document before start transaction test. If not,
    # then a transaction will be auto created inside the active document if a
    # new transaction is triggered from a non active document
    FreeCAD.setActiveDocument('Observer2')
    self.assertEqual(self.Obs.signal.pop(), 'DocActivated')
    self.assertTrue(self.Obs.parameter.pop() is self.Doc2)
    self.assertTrue(not self.Obs.signal and not self.Obs.parameter and not self.Obs.parameter2)

    self.Doc2.openTransaction('test')
    # openTransaction() now only setup pending transaction, which will only be
    # created when there is actual change
    self.Doc2.addObject('App::FeatureTest','test')
    self.assertEqual(self.Obs.signal[0], 'DocOpenTransaction')
    self.assertEqual(self.Obs.signal.count('DocOpenTransaction'),1)
    self.assertTrue(self.Obs.parameter[0] is self.Doc2)
    self.assertEqual(self.Obs.parameter2[0], 'test')
    self.Obs.signal = []
    self.Obs.parameter = []
    self.Obs.parameter2 = []
    
    self.Doc2.commitTransaction()
    self.assertEqual(self.Obs.signal.pop(), 'DocCommitTransaction')
    self.assertTrue(self.Obs.parameter.pop() is self.Doc2)
    self.assertTrue(not self.Obs.signal and not self.Obs.parameter and not self.Obs.parameter2)
    
    self.Doc2.openTransaction('test2')
    # openTransaction() now only setup pending transaction, which will only be
    # created when there is actual change
    self.Doc2.addObject('App::FeatureTest','test')
    self.assertEqual(self.Obs.signal[0], 'DocOpenTransaction')
    self.assertEqual(self.Obs.signal.count('DocOpenTransaction'),1)
    self.assertTrue(self.Obs.parameter[0] is self.Doc2)
    self.assertEqual(self.Obs.parameter2[0], 'test2')
    # there will be other signals because of the addObject()
    self.Obs.signal = []
    self.Obs.parameter = []
    self.Obs.parameter2 = []
    
    self.Doc2.abortTransaction()
    self.assertEqual(self.Obs.signal.pop(), 'DocAbortTransaction')
    self.assertTrue(self.Obs.parameter.pop() is self.Doc2)
    # there will be other signals because of aborting the above addObject()
    self.Obs.signal = []
    self.Obs.parameter = []
    self.Obs.parameter2 = []
    
    self.Doc2.undo()
    self.assertEqual(self.Obs.signal.pop(), 'DocUndo')
    self.assertTrue(self.Obs.parameter.pop() is self.Doc2)
    # there will be other signals because undoing the above addObject()
    self.Obs.signal = []
    self.Obs.parameter = []
    self.Obs.parameter2 = []
    
    self.Doc2.redo()
    self.assertEqual(self.Obs.signal.pop(), 'DocRedo')
    self.assertTrue(self.Obs.parameter.pop() is self.Doc2)
    # there will be other signals because redoing the above addObject()
    self.Obs.signal = []
    self.Obs.parameter = []
    self.Obs.parameter2 = []
    
    self.Doc1.Comment = 'test comment'
    self.assertEqual(self.Obs.signal.pop(0), 'DocBeforeChange')
    self.assertTrue(self.Obs.parameter.pop(0) is self.Doc1)
    self.assertEqual(self.Obs.parameter2.pop(0), 'Comment')
    self.assertEqual(self.Obs.signal.pop(0), 'DocChanged')
    self.assertTrue(self.Obs.parameter.pop(0) is self.Doc1)
    self.assertEqual(self.Obs.parameter2.pop(0), 'Comment')
    
    FreeCAD.closeDocument(self.Doc2.Name)
    self.assertEqual(self.Obs.signal.pop(), 'DocDeleted')
    self.assertTrue(self.Obs.parameter.pop() is self.Doc2)
    if FreeCAD.GuiUp:
        # only has document activated signal when running in GUI mode
        self.assertEqual(self.Obs.signal.pop(), 'DocActivated')
        self.assertTrue(self.Obs.parameter.pop() is self.Doc1)
    self.assertTrue(not self.Obs.signal and not self.Obs.parameter and not self.Obs.parameter2)

    FreeCAD.closeDocument(self.Doc1.Name)
    self.assertEqual(self.Obs.signal.pop(), 'DocDeleted')
    self.assertEqual(self.Obs.parameter.pop(), self.Doc1)
    self.assertTrue(not self.Obs.signal and not self.Obs.parameter and not self.Obs.parameter2)
       
  def testObject(self):
    #testing signal on object changes
    
    self.Doc1 = FreeCAD.newDocument("Observer1")
    self.Obs.signal = []
    self.Obs.parameter = []
    self.Obs.parameter2 = []
    
    obj = self.Doc1.addObject("App::DocumentObject","obj")
    self.failUnless(self.Obs.signal.pop() == 'ObjCreated')
    self.failUnless(self.Obs.parameter.pop() is obj)
    #there are multiple object change signals
    self.Obs.signal = []
    self.Obs.parameter = []
    self.Obs.parameter2 = []
    
    obj.Label = "myobj"
    self.failUnless(self.Obs.signal.pop(0) == 'ObjBeforeChange')
    self.failUnless(self.Obs.parameter.pop(0) is obj)
    self.failUnless(self.Obs.parameter2.pop(0) == "Label")
    self.failUnless(self.Obs.signal.pop(0) == 'ObjChanged')
    self.failUnless(self.Obs.parameter.pop(0) is obj)
    self.failUnless(self.Obs.parameter2.pop(0) == "Label")
    self.failUnless(not self.Obs.signal and not self.Obs.parameter and not self.Obs.parameter2)
    
    obj.enforceRecompute()
    obj.recompute()
    self.failUnless(self.Obs.signal.pop(0) == 'ObjRecomputed')
    self.failUnless(self.Obs.parameter.pop(0) is obj)
    self.failUnless(not self.Obs.signal and not self.Obs.parameter and not self.Obs.parameter2)
    
    obj.enforceRecompute()
    self.Doc1.recompute()
    self.failUnless(self.Obs.signal.pop(0) == 'ObjRecomputed')
    self.failUnless(self.Obs.parameter.pop(0) is obj)
    self.failUnless(self.Obs.signal.pop(0) == 'DocRecomputed')
    self.failUnless(self.Obs.parameter.pop(0) is self.Doc1)
    self.failUnless(not self.Obs.signal and not self.Obs.parameter and not self.Obs.parameter2)
    
    FreeCAD.ActiveDocument.removeObject(obj.Name)
    self.failUnless(self.Obs.signal.pop(0) == 'ObjDeleted')
    self.failUnless(self.Obs.parameter.pop(0) is obj)
    self.failUnless(not self.Obs.signal and not self.Obs.parameter and not self.Obs.parameter2)
    
    pyobj = self.Doc1.addObject("App::FeaturePython","pyobj")
    self.Obs.signal = []
    self.Obs.parameter = []
    self.Obs.parameter2 = []
    pyobj.addProperty("App::PropertyLength","Prop","Group","test property")
    self.failUnless(self.Obs.signal.pop() == 'ObjAddDynProp')
    self.failUnless(self.Obs.parameter.pop() is pyobj)
    self.failUnless(self.Obs.parameter2.pop() == 'Prop')
    self.failUnless(not self.Obs.signal and not self.Obs.parameter and not self.Obs.parameter2)
    
    pyobj.setEditorMode('Prop', ['ReadOnly'])
    self.failUnless(self.Obs.signal.pop() == 'ObjChangePropEdit')
    self.failUnless(self.Obs.parameter.pop() is pyobj)
    self.failUnless(self.Obs.parameter2.pop() == 'Prop')
    self.failUnless(not self.Obs.signal and not self.Obs.parameter and not self.Obs.parameter2)
    
    pyobj.removeProperty('Prop')
    self.failUnless(self.Obs.signal.pop() == 'ObjRemoveDynProp')
    self.failUnless(self.Obs.parameter.pop() is pyobj)
    self.failUnless(self.Obs.parameter2.pop() == 'Prop')
    self.failUnless(not self.Obs.signal and not self.Obs.parameter and not self.Obs.parameter2)
    
    pyobj.addExtension("App::GroupExtensionPython", None)
    self.failUnless(self.Obs.signal.pop() == 'ObjDynExt')
    self.failUnless(self.Obs.parameter.pop() is pyobj)
    self.failUnless(self.Obs.parameter2.pop() == 'App::GroupExtensionPython')
    self.failUnless(self.Obs.signal.pop(0) == 'ObjBeforeDynExt')
    self.failUnless(self.Obs.parameter.pop(0) is pyobj)
    self.failUnless(self.Obs.parameter2.pop(0) == 'App::GroupExtensionPython')
    #a proxy property was changed, hence those events are also in the signal list 
    self.Obs.signal = []
    self.Obs.parameter = []
    self.Obs.parameter2 = []
    
    FreeCAD.closeDocument(self.Doc1.Name)
    self.Obs.signal = []
    self.Obs.parameter = []
    self.Obs.parameter2 = []
    
  def testUndoDisabledDocument(self):

    # testing document level signals
    self.Doc1 = FreeCAD.newDocument("Observer1"); 
    self.Doc1.UndoMode = 0
    self.Obs.signal = []
    self.Obs.parameter = []
    self.Obs.parameter2 = []
     
    self.Doc1.openTransaction('test')  
    self.Doc1.commitTransaction()
    self.Doc1.undo()
    self.Doc1.redo()    
    self.failUnless(not self.Obs.signal and not self.Obs.parameter and not self.Obs.parameter2)
  
    FreeCAD.closeDocument(self.Doc1.Name)
    self.Obs.signal = []
    self.Obs.parameter = []
    self.Obs.parameter2 = []
    
  def testGuiObserver(self):
  
    if not FreeCAD.GuiUp:
      return

    # in case another document already exists then the tests cannot
    # be done reliably
    if FreeCAD.activeDocument():
      return

    self.GuiObs = self.GuiObserver()
    FreeCAD.Gui.addDocumentObserver(self.GuiObs)
    self.Doc1 = FreeCAD.newDocument("Observer1");
    self.GuiDoc1 = FreeCAD.Gui.getDocument(self.Doc1.Name)
    self.Obs.signal = []
    self.Obs.parameter = []
    self.Obs.parameter2 = []
    self.failUnless(self.GuiObs.signal.pop(0) == 'DocCreated')
    self.failUnless(self.GuiObs.parameter.pop(0) is self.GuiDoc1)
    self.failUnless(self.GuiObs.signal.pop(0) == 'DocActivated')
    self.failUnless(self.GuiObs.parameter.pop(0) is self.GuiDoc1)
    self.failUnless(self.GuiObs.signal.pop(0) == 'DocRelabled')
    self.failUnless(self.GuiObs.parameter.pop(0) is self.GuiDoc1)
    self.failUnless(not self.GuiObs.signal and not self.GuiObs.parameter and not self.GuiObs.parameter2)
    
    self.Doc1.Label = "test"
    self.failUnless(self.Obs.signal.pop() == 'DocRelabled')
    self.failUnless(self.Obs.parameter.pop() is self.Doc1)
    #not interested in the change signals
    self.Obs.signal = []
    self.Obs.parameter = []
    self.Obs.parameter2 = []
    self.failUnless(self.GuiObs.signal.pop(0) == 'DocRelabled')
    self.failUnless(self.GuiObs.parameter.pop(0) is self.GuiDoc1)
    self.failUnless(not self.GuiObs.signal and not self.GuiObs.parameter and not self.GuiObs.parameter2)
    
    FreeCAD.setActiveDocument(self.Doc1.Name)
    self.failUnless(self.Obs.signal.pop() == 'DocActivated')
    self.failUnless(self.Obs.parameter.pop() is self.Doc1)
    self.failUnless(not self.Obs.signal and not self.Obs.parameter and not self.Obs.parameter2)
    self.failUnless(self.GuiObs.signal.pop() == 'DocActivated')
    self.failUnless(self.GuiObs.parameter.pop() is self.GuiDoc1)
    self.failUnless(not self.GuiObs.signal and not self.GuiObs.parameter and not self.GuiObs.parameter2)
    
    obj = self.Doc1.addObject("App::FeaturePython","obj")
    self.failUnless(self.Obs.signal.pop() == 'ObjCreated')
    self.failUnless(self.Obs.parameter.pop() is obj)
    #there are multiple object change signals
    self.Obs.signal = []
    self.Obs.parameter = []
    self.Obs.parameter2 = []    
    self.failUnless(self.GuiObs.signal.pop() == "ObjCreated")
    self.failUnless(self.GuiObs.parameter.pop() is obj.ViewObject)

    # There are object change signals, caused by sync of obj.Visibility. Same below.
    self.GuiObs.signal = []
    self.GuiObs.parameter = []
    self.GuiObs.parameter2 = []

    obj.ViewObject.Visibility = False
    self.failUnless(self.Obs.signal.pop() == "ObjChanged")
    self.failUnless(self.Obs.parameter.pop() is obj)
    self.failUnless(self.Obs.parameter2.pop() == "Visibility")
    self.failUnless(self.Obs.signal.pop() == "ObjBeforeChange")
    self.failUnless(self.Obs.parameter.pop() is obj)
    self.failUnless(self.Obs.parameter2.pop() == "Visibility")
    self.failUnless(not self.Obs.signal and not self.Obs.parameter and not self.Obs.parameter2)
    self.failUnless(self.GuiObs.signal.pop(0) == 'ObjChanged')
    self.failUnless(self.GuiObs.parameter.pop(0) is obj.ViewObject)
    self.failUnless(self.GuiObs.parameter2.pop(0) == "Visibility")
    self.failUnless(not self.GuiObs.signal and not self.GuiObs.parameter and not self.GuiObs.parameter2)
    
    obj.ViewObject.addProperty("App::PropertyLength","Prop","Group","test property")
    self.failUnless(self.Obs.signal.pop() == 'ObjAddDynProp')
    self.failUnless(self.Obs.parameter.pop() is obj.ViewObject)
    self.failUnless(self.Obs.parameter2.pop() == 'Prop')
    self.failUnless(not self.Obs.signal and not self.Obs.parameter and not self.Obs.parameter2)
    self.failUnless(not self.GuiObs.signal and not self.GuiObs.parameter and not self.GuiObs.parameter2)
    
    obj.ViewObject.setEditorMode('Prop', ['ReadOnly'])
    self.failUnless(self.Obs.signal.pop() == 'ObjChangePropEdit')
    self.failUnless(self.Obs.parameter.pop() is obj.ViewObject)
    self.failUnless(self.Obs.parameter2.pop() == 'Prop')
    self.failUnless(not self.Obs.signal and not self.Obs.parameter and not self.Obs.parameter2)
    self.failUnless(not self.GuiObs.signal and not self.GuiObs.parameter and not self.GuiObs.parameter2)
    
    obj.ViewObject.removeProperty('Prop')
    self.failUnless(self.Obs.signal.pop() == 'ObjRemoveDynProp')
    self.failUnless(self.Obs.parameter.pop() is obj.ViewObject)
    self.failUnless(self.Obs.parameter2.pop() == 'Prop')
    self.failUnless(not self.Obs.signal and not self.Obs.parameter and not self.Obs.parameter2)
    self.failUnless(not self.GuiObs.signal and not self.GuiObs.parameter and not self.GuiObs.parameter2)

    self.GuiDoc1.setEdit('obj', 0)
    self.failUnless(not self.Obs.signal and not self.Obs.parameter and not self.Obs.parameter2)
    self.failUnless(self.GuiObs.signal.pop(0) == 'ObjInEdit')
    self.failUnless(self.GuiObs.parameter.pop(0) is obj.ViewObject)
    self.failUnless(not self.GuiObs.signal and not self.GuiObs.parameter and not self.GuiObs.parameter2)

    self.GuiDoc1.resetEdit()
    self.failUnless(not self.Obs.signal and not self.Obs.parameter and not self.Obs.parameter2)
    self.failUnless(self.GuiObs.signal.pop(0) == 'ObjResetEdit')
    self.failUnless(self.GuiObs.parameter.pop(0) is obj.ViewObject)
    self.failUnless(not self.GuiObs.signal and not self.GuiObs.parameter and not self.GuiObs.parameter2)
    
    obj.ViewObject.addExtension("Gui::ViewProviderGroupExtensionPython", None)
    self.failUnless(self.Obs.signal.pop() == 'ObjDynExt')
    self.failUnless(self.Obs.parameter.pop() is obj.ViewObject)
    self.failUnless(self.Obs.parameter2.pop() == 'Gui::ViewProviderGroupExtensionPython')
    self.failUnless(self.Obs.signal.pop() == 'ObjBeforeDynExt')
    self.failUnless(self.Obs.parameter.pop() is obj.ViewObject)
    self.failUnless(self.Obs.parameter2.pop() == 'Gui::ViewProviderGroupExtensionPython')
    #a proxy property was changed, hence those events are also in the signal list (but of GUI observer)
    self.GuiObs.signal = []
    self.GuiObs.parameter = []
    self.GuiObs.parameter2 = []
    
    vo = obj.ViewObject
    FreeCAD.ActiveDocument.removeObject(obj.Name)
    self.failUnless(self.Obs.signal.pop(0) == 'ObjDeleted')
    self.failUnless(self.Obs.parameter.pop(0) is obj)
    self.failUnless(not self.Obs.signal and not self.Obs.parameter and not self.Obs.parameter2)
    self.failUnless(self.GuiObs.signal.pop() == 'ObjDeleted')
    self.failUnless(self.GuiObs.parameter.pop() is vo)
    self.failUnless(not self.GuiObs.signal and not self.GuiObs.parameter and not self.GuiObs.parameter2)

    FreeCAD.closeDocument(self.Doc1.Name)
    self.Obs.signal = []
    self.Obs.parameter = []
    self.Obs.parameter2 = []
    self.failUnless(self.GuiObs.signal.pop() == 'DocDeleted')
    self.failUnless(self.GuiObs.parameter.pop() is self.GuiDoc1)
    self.failUnless(not self.GuiObs.signal and not self.GuiObs.parameter and not self.GuiObs.parameter2)

    FreeCAD.Gui.removeDocumentObserver(self.GuiObs)

  def tearDown(self):
    #closing doc
    FreeCAD.removeDocumentObserver(self.Obs)
    self.Obs = None



# -*- coding: utf-8 -*-

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2014 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

from __future__ import print_function

import FreeCAD
import FreeCADGui
import Path
import PathScripts
import PathScripts.PathLog as PathLog
import PathScripts.PathPreferences as PathPreferences
import PathScripts.PathToolBitLibraryCmd as PathToolBitLibraryCmd
import PathScripts.PathToolEdit as PathToolEdit
import PathScripts.PathUtils as PathUtils
import PathScripts.PathToolLibraryManager as ToolLibraryManager

from PySide import QtCore, QtGui

PathLog.setLevel(PathLog.Level.INFO, PathLog.thisModule())
#PathLog.trackModule(PathLog.thisModule())

def translate(context, text, disambig=None):
    return QtCore.QCoreApplication.translate(context, text, disambig)

class EditorPanel():

    def __init__(self, job, cb):
        self.form = FreeCADGui.PySideUic.loadUi(":/panels/ToolLibraryEditor.ui")
        self.TLM = ToolLibraryManager.ToolLibraryManager()
        listname = self.TLM.getCurrentTableName()
        
        if listname:
            self.loadToolTables()

        self.job = job
        self.cb = cb

    def toolEditor(self, tool):
        dialog = FreeCADGui.PySideUic.loadUi(":/panels/DlgToolEdit.ui")
        editor = PathToolEdit.ToolEditor(tool, dialog.toolEditor, dialog)
        editor.setupUI()
        return editor

    def accept(self):
        pass

    def reject(self):
        FreeCADGui.Control.closeDialog()
        FreeCAD.ActiveDocument.recompute()

    def getFields(self):
        pass

    def setFields(self):
        pass

    def open(self):
        pass

    def getType(self, tooltype):
        "gets a combobox index number for a given type or viceversa"
        toolslist = Path.Tool.getToolTypes(Path.Tool())
        if isinstance(tooltype, str):
            if tooltype in toolslist:
                return toolslist.index(tooltype)
            else:
                return 0
        else:
            return toolslist[tooltype]

    def getMaterial(self, material):
        '''gets a combobox index number for a given material or viceversa'''
        matslist = Path.Tool.getToolMaterials(Path.Tool())
        if isinstance(material, str):
            if material in matslist:
                return matslist.index(material)
            else:
                return 0
        else:
            return matslist[material]

    def addTool(self):
        '''adds new tool to the current tool table'''
        tool = Path.Tool()
        editor = self.toolEditor(tool)

        r = editor.Parent.exec_()
        if r:
            editor.accept()
            listname = self.TLM.getCurrentTableName()
            self.TLM.addnew(listname, editor.Tool)
            self.loadTable(listname)

    def delete(self):
        '''deletes the selected tool'''
        listname = self.TLM.getCurrentTableName()
        model = self.form.ToolsList.model()
        for i in range(model.rowCount()):
            item = model.item(i, 0)
            if item.checkState():
                t = model.index(i, 1)
                self.TLM.delete(int(t.data()) ,listname)
        self.loadTable(listname)
        self.toolSelectionChanged()

    def editTool(self, currItem):
        '''load the tool edit dialog'''
        if not currItem:
            currItem = self.form.ToolsList.selectedIndexes()[1]

        row = currItem.row()
        value = currItem.sibling(row, 1).data()
        listname = self.TLM.getCurrentTableName()
        toolnum = int(value)
        tool = self.TLM.getTool(listname, toolnum)
        editor = self.toolEditor(tool)

        r = editor.Parent.exec_()
        if r:
            editor.accept()
            if self.TLM.updateTool(listname, toolnum, editor.Tool) is True:
                self.loadTable(listname)

    def moveUp(self):
        '''moves a tool to a lower number, if possible'''
        item = self.form.ToolsList.selectedIndexes()[1].data()
        if item:
            number = int(item)
            listname = self.TLM.getCurrentTableName()
            success, newNum = self.TLM.moveup(number, listname)
            if success:
                self.loadTable(listname)
                self.updateSelection(newNum)

    def moveDown(self):
        '''moves a tool to a higher number, if possible'''
        item = self.form.ToolsList.selectedIndexes()[1].data()
        if item:
            number = int(item)
            listname = self.TLM.getCurrentTableName()
            success, newNum = self.TLM.movedown(number, listname) 
            if success:
                self.loadTable(listname)
                self.updateSelection(newNum)

    def duplicate(self):
        '''duplicated the selected tool in the current tool table'''
        item = self.form.ToolsList.selectedIndexes()[1].data()
        if item:
            number = int(item)
            listname = self.TLM.getCurrentTableName()
            success, newNum = self.TLM.duplicate(number, listname)
            if success:
                self.loadTable(listname)
                self.updateSelection(newNum)

    def updateSelection(self, number):
        '''update the tool list selection to track moves'''
        model = self.form.ToolsList.model()
        for i in range(model.rowCount()):
            if int(model.index(i, 1).data()) == number:
                self.form.ToolsList.selectRow(i)
                self.form.ToolsList.model().item(i, 0).setCheckState(QtCore.Qt.Checked)
                return


    def importFile(self):
        '''imports a tooltable from a file'''
        filename = QtGui.QFileDialog.getOpenFileName(self.form, translate( "TooltableEditor", "Open tooltable", None), None, "{};;{};;{}".format(self.TLM.TooltableTypeJSON, self.TLM.TooltableTypeXML, self.TLM.TooltableTypeHeekscad))
        if filename[0]:
            listname = self.TLM.getNextToolTableName()
            if self.TLM.read(filename, listname):
                self.loadToolTables()

    def exportFile(self):
        '''export a tooltable to a file'''
        filename = QtGui.QFileDialog.getSaveFileName(self.form, translate("TooltableEditor", "Save tooltable", None), None, "{};;{};;{}".format(self.TLM.TooltableTypeJSON, self.TLM.TooltableTypeXML, self.TLM.TooltableTypeLinuxCNC))
        if filename[0]:
            listname = self.TLM.getCurrentTableName()
            self.TLM.write(filename, listname)

    def toolSelectionChanged(self, index=None):
        ''' updates the ui when tools are selected'''
        if index:
            self.form.ToolsList.selectRow(index.row())
        
        self.form.btnCopyTools.setEnabled(False)
        self.form.ButtonDelete.setEnabled(False)
        self.form.ButtonUp.setEnabled(False)
        self.form.ButtonDown.setEnabled(False)
        self.form.ButtonEdit.setEnabled(False)
        self.form.ButtonDuplicate.setEnabled(False)

        model = self.form.ToolsList.model()
        checkCount = 0
        checkList = []
        for i in range(model.rowCount()):
            item = model.item(i, 0)
            if item.checkState():
                checkCount += 1
                checkList.append(i)
                self.form.btnCopyTools.setEnabled(True)

        # only allow moving or deleting a single tool at a time.
        if checkCount == 1:
            #make sure the row is highlighted when the check box gets ticked   
            self.form.ToolsList.selectRow(checkList[0])       
            self.form.ButtonDelete.setEnabled(True)
            self.form.ButtonUp.setEnabled(True)
            self.form.ButtonDown.setEnabled(True)
            self.form.ButtonEdit.setEnabled(True)
            self.form.ButtonDuplicate.setEnabled(True)

        if len(PathUtils.GetJobs()) == 0:
            self.form.btnCopyTools.setEnabled(False)

    def copyTools(self):
        ''' copy selected tool '''
        tools = []
        model = self.form.ToolsList.model()
        for i in range(model.rowCount()):
            item = model.item(i, 0)
            if item.checkState():
                item = model.index(i, 1)
                tools.append(item.data())
        if len(tools) == 0:
            return

        targets = self.TLM.getJobList()
        currList = self.TLM.getCurrentTableName()

        for target in targets:
            if target == currList:
                targets.remove(target)

        if len(targets) == 0:
            FreeCAD.Console.PrintWarning("No Path Jobs in current document")
            return
        elif len(targets) == 1:
            targetlist = targets[0]
        else:
            form = FreeCADGui.PySideUic.loadUi(":/panels/DlgToolCopy.ui")
            form.cboTarget.addItems(targets)
            r = form.exec_()
            if r is False:
                return None
            else:
                targetlist = form.cboTarget.currentText()

        for toolnum in tools:
            tool = self.TLM.getTool(currList, int(toolnum))
            PathLog.debug('tool: {}, toolnum: {}'.format(tool, toolnum))
            if self.job:
                label = "T{}: {}".format(toolnum, tool.Name)
                tc = PathScripts.PathToolController.Create(label, tool=tool, toolNumber=int(toolnum))
                self.job.Proxy.addToolController(tc)
            else:
                for job in FreeCAD.ActiveDocument.findObjects("Path::Feature"):
                    if isinstance(job.Proxy, PathScripts.PathJob.ObjectJob) and job.Label == targetlist:
                        label = "T{}: {}".format(toolnum, tool.Name)
                        tc = PathScripts.PathToolController.Create(label, tool=tool, toolNumber=int(toolnum))
                        job.Proxy.addToolController(tc)
        if self.cb:
            self.cb()
        FreeCAD.ActiveDocument.recompute()

    def tableSelected(self, index):
        ''' loads the tools for the selected tool table '''
        name = self.form.TableList.itemWidget(self.form.TableList.itemFromIndex(index)).getTableName()
        self.loadTable(name)

    def loadTable(self, name):
        ''' loads the tools for the selected tool table '''
        tooldata = self.TLM.getTools(name)
        if tooldata:
            self.form.ToolsList.setModel(tooldata)
            self.form.ToolsList.resizeColumnsToContents()
            self.form.ToolsList.horizontalHeader().setResizeMode(self.form.ToolsList.model().columnCount() - 1, QtGui.QHeaderView.Stretch)       
            self.setCurrentToolTableByName(name)

    def addNewToolTable(self):
        ''' adds new tool to selected tool table '''
        name = self.TLM.addNewToolTable()
        self.loadToolTables()
        self.loadTable(name)
         
    def loadToolTables(self):
        ''' Load list of available tool tables '''
        self.form.TableList.clear()
        model = self.form.ToolsList.model()
        if model:
            model.clear()
        if len(self.TLM.getToolTables()) > 0:
            for table in self.TLM.getToolTables():
                listWidgetItem = QtGui.QListWidgetItem()
                listItem = ToolTableListWidgetItem(self.TLM)
                listItem.setTableName(table.Name)
                listItem.setIcon(QtGui.QPixmap(':/icons/Path-ToolTable.svg'))
                listItem.toolMoved.connect(self.reloadReset)
                listWidgetItem.setSizeHint(QtCore.QSize(0,40))
                self.form.TableList.addItem(listWidgetItem)
                self.form.TableList.setItemWidget(listWidgetItem, listItem)
            #Load the first tooltable
            self.loadTable(self.TLM.getCurrentTableName())

    def reloadReset(self):
        ''' reloads the current tooltable'''
        name = self.TLM.getCurrentTableName()
        self.loadTable(name)
        
    def setCurrentToolTableByName(self, name):
        ''' get the current tool table '''
        item = self.getToolTableByName(name)
        if item:
            self.form.TableList.setCurrentItem(item)

    def getToolTableByName(self, name):
        ''' returns the listWidgetItem for the selected name'''
        for i in range(self.form.TableList.count()):
            tableName = self.form.TableList.itemWidget(self.form.TableList.item(i)).getTableName()
            if tableName == name:
                return self.form.TableList.item(i)
        return False

    def removeToolTable(self):
        ''' delete the selected tool table '''
        self.TLM.deleteToolTable()
        self.loadToolTables()
    
    def renameTable(self):
        ''' provides dialog for new tablename and renames the selected tool table'''
        name = self.TLM.getCurrentTableName()
        newName, ok = QtGui.QInputDialog.getText(None, translate("TooltableEditor","Rename Tooltable"),translate("TooltableEditor","Enter Name:"),QtGui.QLineEdit.Normal,name)
        if ok and newName:
            index = self.form.TableList.indexFromItem(self.getToolTableByName(name)).row()
            reloadTables = self.TLM.renameToolTable(newName, index)
            if reloadTables:
                self.loadToolTables()
                self.loadTable(newName)

    def getStandardButtons(self):
        return int(QtGui.QDialogButtonBox.Ok)

    def setupUi(self):
        # Connect Signals and Slots
        self.form.ButtonNewTool.clicked.connect(self.addTool)
        self.form.ButtonImport.clicked.connect(self.importFile)
        self.form.ButtonExport.clicked.connect(self.exportFile)
        self.form.ButtonDown.clicked.connect(self.moveDown)
        self.form.ButtonUp.clicked.connect(self.moveUp)
        self.form.ButtonDelete.clicked.connect(self.delete)
        self.form.ButtonEdit.clicked.connect(self.editTool)
        self.form.ButtonDuplicate.clicked.connect(self.duplicate)
        self.form.btnCopyTools.clicked.connect(self.copyTools)
 
        self.form.ToolsList.doubleClicked.connect(self.editTool)
        self.form.ToolsList.clicked.connect(self.toolSelectionChanged)
 
        self.form.TableList.clicked.connect(self.tableSelected)
        self.form.TableList.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.form.TableList.itemChanged.connect(self.renameTable)

        self.form.ButtonAddToolTable.clicked.connect(self.addNewToolTable)
        self.form.ButtonAddToolTable.setToolTip(translate("TooltableEditor","Add New Tool Table"))
        self.form.ButtonRemoveToolTable.clicked.connect(self.removeToolTable)
        self.form.ButtonRemoveToolTable.setToolTip(translate("TooltableEditor","Delete Selected Tool Table"))
        self.form.ButtonRenameToolTable.clicked.connect(self.renameTable)
        self.form.ButtonRenameToolTable.setToolTip(translate("TooltableEditor","Rename Selected Tool Table"))

        self.setFields()

class ToolTableListWidgetItem(QtGui.QWidget):
    
    toolMoved = QtCore.Signal()
  
    def __init__(self, TLM):
        super(ToolTableListWidgetItem, self).__init__()
        
        self.tlm = TLM   
        self.setAcceptDrops(True)

        self.mainLayout = QtGui.QHBoxLayout()
        self.iconQLabel = QtGui.QLabel()
        self.tableNameLabel = QtGui.QLabel()
        self.mainLayout.addWidget(self.iconQLabel, 0)
        self.mainLayout.addWidget(self.tableNameLabel, 1)
        self.setLayout(self.mainLayout)

    def setTableName (self, text):
        self.tableNameLabel.setText(text)

    def getTableName(self):
        return self.tableNameLabel.text()

    def setIcon (self, icon):
        icon = icon.scaled(24, 24)
        self.iconQLabel.setPixmap(icon)

    def dragEnterEvent(self, e):
        currentToolTable = self.tlm.getCurrentTableName()
        thisToolTable = self.getTableName()
      
        if not currentToolTable == thisToolTable:
            e.accept()
        else:
            e.ignore() 

    def dropEvent(self, e):
        selectedTools = e.source().selectedIndexes()
        if selectedTools:
            toolData = selectedTools[1].data()
        
            if toolData:
                self.tlm.moveToTable(int(toolData), self.getTableName())
                self.toolMoved.emit()


class CommandToolLibraryEdit():
    def __init__(self):
        pass

    def edit(self, job=None, cb=None):
        if PathPreferences.toolsReallyUseLegacyTools():
            editor = EditorPanel(job, cb)
            editor.setupUi()
            editor.form.exec_()
        else:
            if PathToolBitLibraryCmd.CommandToolBitLibraryLoad.Execute(job):
                if cb:
                    cb()

    def GetResources(self):
        return {'Pixmap'  : 'Path-ToolTable',
                'MenuText': QtCore.QT_TRANSLATE_NOOP("Path_ToolTable","Tool Manager"),
                'Accel': "P, T",
                'ToolTip': QtCore.QT_TRANSLATE_NOOP("Path_ToolTable","Tool Manager")}

    def IsActive(self):
        return not FreeCAD.ActiveDocument is None

    def Activated(self):
        self.edit()

if FreeCAD.GuiUp:
    # register the FreeCAD command
    FreeCADGui.addCommand('Path_ToolLibraryEdit',CommandToolLibraryEdit())



# -*- coding: utf8 -*-

#***************************************************************************
#*   Copyright (c) 2018 Yorik van Havre <yorik@uncreated.net>              *
#*                                                                         *
#*   This program is free software; you can redistribute it and/or modify  *
#*   it under the terms of the GNU Lesser General Public License (LGPL)    *
#*   as published by the Free Software Foundation; either version 2 of     *
#*   the License, or (at your option) any later version.                   *
#*   for detail see the LICENCE text file.                                 *
#*                                                                         *
#*   This program is distributed in the hope that it will be useful,       *
#*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
#*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
#*   GNU Library General Public License for more details.                  *
#*                                                                         *
#*   You should have received a copy of the GNU Library General Public     *
#*   License along with this program; if not, write to the Free Software   *
#*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
#*   USA                                                                   *
#*                                                                         *
#***************************************************************************

import FreeCAD
import Draft
import ArchCommands
import DraftVecUtils
import sys
import ArchIFC
import tempfile
import os
if FreeCAD.GuiUp:
    import FreeCADGui
    from PySide import QtCore, QtGui
    from DraftTools import translate
    from PySide.QtCore import QT_TRANSLATE_NOOP
else:
    # \cond
    def translate(ctxt,txt):
        return txt
    def QT_TRANSLATE_NOOP(ctxt,txt):
        return txt
    # \endcond
if sys.version_info.major >= 3:
    unicode = str

## @package ArchBuildingPart
#  \ingroup ARCH
#  \brief The BuildingPart object and tools
#
#  This module provides tools to build BuildingPart objects.
#  BuildingParts are used to group different Arch objects

__title__="FreeCAD Arch BuildingPart"
__author__ = "Yorik van Havre"
__url__ = "http://www.freecadweb.org"


BuildingTypes = ['Undefined',
'Agricultural - Barn',
'Agricultural - Chicken coop or chickenhouse',
'Agricultural - Cow-shed',
'Agricultural - Farmhouse',
'Agricultural - Granary',
'Agricultural - Greenhouse',
'Agricultural - Hayloft',
'Agricultural - Pigpen or sty',
'Agricultural - Root cellar',
'Agricultural - Shed',
'Agricultural - Silo',
'Agricultural - Stable',
'Agricultural - Storm cellar',
'Agricultural - Well house',
'Agricultural - Underground pit',

'Commercial - Automobile repair shop',
'Commercial - Bank',
'Commercial - Car wash',
'Commercial - Convention center',
'Commercial - Forum',
'Commercial - Gas station',
'Commercial - Hotel',
'Commercial - Market',
'Commercial - Market house',
'Commercial - Skyscraper',
'Commercial - Shop',
'Commercial - Shopping mall',
'Commercial - Supermarket',
'Commercial - Warehouse',
'Commercial - Restaurant',

'Residential - Apartment block',
'Residential - Asylum',
'Residential - Condominium',
'Residential - Dormitory',
'Residential - Duplex',
'Residential - House',
'Residential - Nursing home',
'Residential - Townhouse',
'Residential - Villa',
'Residential - Bungalow',

'Educational - Archive',
'Educational - College classroom building',
'Educational - College gymnasium',
'Educational - College students union',
'Educational - School',
'Educational - Library',
'Educational - Museum',
'Educational - Art gallery',
'Educational - Theater',
'Educational - Amphitheater',
'Educational - Concert hall',
'Educational - Cinema',
'Educational - Opera house',
'Educational - Boarding school',

'Government - Capitol',
'Government - City hall',
'Government - Consulate',
'Government - Courthouse',
'Government - Embassy',
'Government - Fire station',
'Government - Meeting house',
'Government - Moot hall',
'Government - Palace',
'Government - Parliament',
'Government - Police station',
'Government - Post office',
'Government - Prison',

'Industrial - Brewery',
'Industrial - Factory',
'Industrial - Foundry',
'Industrial - Power plant',
'Industrial - Mill',

'Military - Arsenal',
'Military -Barracks',

'Parking - Boathouse',
'Parking - Garage',
'Parking - Hangar',

'Storage - Silo',
'Storage - Hangar',

'Religious - Church',
'Religious - Basilica',
'Religious - Cathedral',
'Religious - Chapel',
'Religious - Oratory',
'Religious - Martyrium',
'Religious - Mosque',
'Religious - Mihrab',
'Religious - Surau',
'Religious - Imambargah',
'Religious - Monastery',
'Religious - Mithraeum',
'Religious - Fire temple',
'Religious - Shrine',
'Religious - Synagogue',
'Religious - Temple',
'Religious - Pagoda',
'Religious - Gurdwara',
'Religious - Hindu temple',

'Transport - Airport terminal',
'Transport - Bus station',
'Transport - Metro station',
'Transport - Taxi station',
'Transport - Railway station',
'Transport - Signal box',
'Transport - Lighthouse',

'Infrastructure - Data centre',

'Power station - Fossil-fuel power station',
'Power station - Nuclear power plant',
'Power station - Geothermal power',
'Power station - Biomass-fuelled power plant',
'Power station - Waste heat power plant',
'Power station - Renewable energy power station',
'Power station - Atomic energy plant',

'Other - Apartment',
'Other - Clinic',
'Other - Community hall',
'Other - Eatery',
'Other - Folly',
'Other - Food court',
'Other - Hospice',
'Other - Hospital',
'Other - Hut',
'Other - Bathhouse',
'Other - Workshop',
'Other - World trade centre'
]


def makeBuildingPart(objectslist=None,baseobj=None,name="BuildingPart"):

    '''makeBuildingPart(objectslist): creates a buildingPart including the
    objects from the given list.'''

    obj = FreeCAD.ActiveDocument.addObject("App::GeometryPython","BuildingPart")
    #obj = FreeCAD.ActiveDocument.addObject("App::FeaturePython","BuildingPart")
    obj.Label = translate("Arch","BuildingPart")
    BuildingPart(obj)
    # if no IfcType is set it will be the first in the available
    # Annotation in IFC2x3 and Actuator in IFC4, both is certainly wrong
    # use Undefined ATM
    obj.IfcType = "Undefined"
    if FreeCAD.GuiUp:
        ViewProviderBuildingPart(obj.ViewObject)
    if objectslist:
        obj.addObjects(objectslist)
    return obj


def makeFloor(objectslist=None,baseobj=None,name="Floor"):

    """overwrites ArchFloor.makeFloor"""

    obj = makeBuildingPart(objectslist)
    obj.Label = name
    obj.IfcType = "Building Storey"
    return obj


def makeBuilding(objectslist=None,baseobj=None,name="Building"):

    """overwrites ArchBuilding.makeBuilding"""

    obj = makeBuildingPart(objectslist)
    obj.Label = name
    obj.IfcType = "Building"
    obj.addProperty("App::PropertyEnumeration","BuildingType","Building",QT_TRANSLATE_NOOP("App::Property","The type of this building"))
    obj.BuildingType = BuildingTypes
    if FreeCAD.GuiUp:
        obj.ViewObject.ShowLevel = False
        obj.ViewObject.ShowLabel = False
    return obj


def convertFloors(floor=None):

    """convert the given Floor or Building (or all Arch Floors from the active document if none is given) into BuildingParts"""

    todel = []
    if floor:
        objset = [floor]
    else:
        objset = FreeCAD.ActiveDocument.Objects
    for obj in objset:
        if Draft.getType(obj) in ["Floor","Building"]:
            nobj = makeBuildingPart(obj.Group)
            if Draft.getType(obj) == "Floor":
                nobj.IfcType = "Building Storey"
            else:
                nobj.IfcType = "Building"
                nobj.addProperty("App::PropertyEnumeration","BuildingType","Building",QT_TRANSLATE_NOOP("App::Property","The type of this building"))
                nobj.BuildingType = BuildingTypes
            label = obj.Label
            for parent in obj.InList:
                if hasattr(parent,"Group"):
                    if obj in parent.Group:
                        parent.addObject(nobj)
                        #g = parent.Group
                        #g.append(nobj)
                        #parent.Group = g
                else:
                    print("Warning: couldn't add new object '"+label+"' to parent object '"+parent.Label+"'")
            todel.append(obj.Name)
            if obj.ViewObject:
                obj.ViewObject.Proxy.Object = None # some bug makes this trigger even efter the object has been deleted...
            obj.Label = obj.Label+" to delete" # in case FreeCAD doesn't allow 2 objs with same label
            nobj.Label = label
    for n in todel:
        from DraftGui import todo
        todo.delay(FreeCAD.ActiveDocument.removeObject,n)



class CommandBuildingPart:


    "the Arch BuildingPart command definition"

    def GetResources(self):

        return {'Pixmap'  : 'Arch_BuildingPart',
                'MenuText': QT_TRANSLATE_NOOP("Arch_BuildingPart","BuildingPart"),
                'Accel': "B, P",
                'ToolTip': QT_TRANSLATE_NOOP("Arch_BuildingPart","Creates a BuildingPart object including selected objects")}

    def IsActive(self):

        return not FreeCAD.ActiveDocument is None

    def Activated(self):

        sel = FreeCADGui.Selection.getSelection()
        ss = "[ "
        for o in sel:
            ss += "FreeCAD.ActiveDocument." + o.Name + ", "
        ss += "]"
        FreeCAD.ActiveDocument.openTransaction(translate("Arch","Create BuildingPart"))
        FreeCADGui.addModule("Arch")
        FreeCADGui.doCommand("obj = Arch.makeBuildingPart("+ss+")")
        FreeCADGui.addModule("Draft")
        FreeCADGui.doCommand("obj.Placement = FreeCAD.DraftWorkingPlane.getPlacement()")
        FreeCADGui.doCommand("Draft.autogroup(obj)")
        FreeCAD.ActiveDocument.commitTransaction()
        FreeCAD.ActiveDocument.recompute()



class BuildingPart(ArchIFC.IfcProduct):


    "The BuildingPart object"

    def __init__(self,obj):

        obj.Proxy = self
        obj.addExtension('App::GroupExtensionPython', self)
        #obj.addExtension('App::OriginGroupExtensionPython', self)
        self.setProperties(obj)

    def setProperties(self,obj):
        ArchIFC.IfcProduct.setProperties(self, obj)

        pl = obj.PropertiesList
        if not "Height" in pl:
            obj.addProperty("App::PropertyLength","Height","BuildingPart",QT_TRANSLATE_NOOP("App::Property","The height of this object"))
        if not "HeightPropagate" in pl:
            obj.addProperty("App::PropertyBool","HeightPropagate","Children",QT_TRANSLATE_NOOP("App::Property","If true, the height value propagates to contained objects"))
            obj.HeightPropagate = True
        if not "LevelOffset" in pl:
            obj.addProperty("App::PropertyLength","LevelOffset","BuildingPart",QT_TRANSLATE_NOOP("App::Property","The level of the (0,0,0) point of this level"))
        if not "Area" in pl:
            obj.addProperty("App::PropertyArea","Area", "BuildingPart",QT_TRANSLATE_NOOP("App::Property","The computed floor area of this floor"))
        if not "Description" in pl:
            obj.addProperty("App::PropertyString","Description","Component",QT_TRANSLATE_NOOP("App::Property","An optional description for this component"))
        if not "Tag" in pl:
            obj.addProperty("App::PropertyString","Tag","Component",QT_TRANSLATE_NOOP("App::Property","An optional tag for this component"))
        if not "Shape" in pl:
            obj.addProperty("Part::PropertyPartShape","Shape","BuildingPart",QT_TRANSLATE_NOOP("App::Property","The shape of this object"))
        if not "SavedInventor" in pl:
            obj.addProperty("App::PropertyFileIncluded","SavedInventor","BuildingPart",QT_TRANSLATE_NOOP("App::Property","This property stores an inventor representation for this object"))
            obj.setEditorMode("SavedInventor",2)

        self.Type = "BuildingPart"

    def onDocumentRestored(self,obj):

        self.setProperties(obj)

    def __getstate__(self):

        return None

    def __setstate__(self,state):

        return None

    def onBeforeChange(self,obj,prop):

        if prop == "Placement":
            self.oldPlacement = FreeCAD.Placement(obj.Placement)

    def onChanged(self,obj,prop):

        ArchIFC.IfcProduct.onChanged(self, obj, prop)

        # clean svg cache if needed
        if prop in ["Placement","Group"]:
            self.svgcache = None
            self.shapecache = None

        if (prop == "Height" or prop == "HeightPropagate") and obj.Height.Value:
            self.touchChildren(obj)

        elif prop == "Placement":
            if hasattr(self,"oldPlacement"):
                if self.oldPlacement and (self.oldPlacement != obj.Placement):
                    deltap = obj.Placement.Base.sub(self.oldPlacement.Base)
                    if deltap.Length == 0:
                        deltap = None
                    v = FreeCAD.Vector(0,0,1)
                    deltar = FreeCAD.Rotation(self.oldPlacement.Rotation.multVec(v),obj.Placement.Rotation.multVec(v))
                    #print "Rotation",deltar.Axis,deltar.Angle
                    if deltar.Angle < 0.0001:
                        deltar = None
                    for child in obj.Group:
                        if ((not hasattr(child,"MoveWithHost")) or child.MoveWithHost) and hasattr(child,"Placement"):
                            #print "moving ",child.Label
                            if deltar:
                                #child.Placement.Rotation = child.Placement.Rotation.multiply(deltar) - not enough, child must also move
                                # use shape methods to obtain a correct placement
                                import Part,math
                                shape = Part.Shape()
                                shape.Placement = child.Placement
                                #print("angle before rotation:",shape.Placement.Rotation.Angle)
                                #print("rotation angle:",math.degrees(deltar.Angle))
                                shape.rotate(DraftVecUtils.tup(obj.Placement.Base), DraftVecUtils.tup(deltar.Axis), math.degrees(deltar.Angle))
                                print("angle after rotation:",shape.Placement.Rotation.Angle)
                                child.Placement = shape.Placement
                            if deltap:
                                print("moving child")
                                child.Placement.move(deltap)

    def execute(self,obj):

        # gather all the child shapes into a compound
        shapes = self.getShapes(obj)
        if shapes:
            f = []
            for s in shapes:
                f.extend(s.Faces)
            #print("faces before compound:",len(f))
            import Part
            obj.Shape = Part.makeCompound(f)
            #print("faces after compound:",len(obj.Shape.Faces))
            #print("recomputing ",obj.Label)
        obj.Area = self.getArea(obj)

    def getArea(self,obj):

        "computes the area of this floor by adding its inner spaces"

        area = 0
        if hasattr(obj,"Group"):
            for child in obj.Group:
                if hasattr(child,"Area") and hasattr(child,"IfcType"):
                    # only add arch objects that have an Area property
                    # TODO only spaces? ATM only spaces and windows have an Area property
                    area += child.Area.Value
        return area

    def getShapes(self,obj):

        "recursively get the shapes of objects inside this BuildingPart"

        shapes = []
        for child in Draft.getGroupContents(obj):
            if hasattr(child,'Shape'):
                shapes.extend(child.Shape.Faces)
        return shapes

    def getSpaces(self,obj):

        "gets the list of Spaces that have this object as their Zone property"

        g = []
        for o in obj.OutList:
            if hasattr(o,"Zone"):
                if o.Zone == obj:
                    g.append(o)
        return g

    def touchChildren(self,obj):
        
        "Touches all descendents where applicable"

        for child in obj.Group:
            if Draft.getType(child) in ["Wall","Structure"]:
                if not child.Height.Value:
                    print("Executing ",child.Label)
                    child.Proxy.execute(child)
            elif Draft.getType(child) in ["Group","BuildingPart"]:
                self.touchChildren(child)


class ViewProviderBuildingPart:


    "A View Provider for the BuildingPart object"

    def __init__(self,vobj):

        vobj.addExtension("Gui::ViewProviderGroupExtensionPython", self)
        #vobj.addExtension("Gui::ViewProviderGeoFeatureGroupExtensionPython", self)
        vobj.Proxy = self
        self.setProperties(vobj)
        vobj.ShapeColor = ArchCommands.getDefaultColor("Helpers")

    def setProperties(self,vobj):

        pl = vobj.PropertiesList
        if not "LineWidth" in pl:
            vobj.addProperty("App::PropertyFloat","LineWidth","BuildingPart",QT_TRANSLATE_NOOP("App::Property","The line width of this object"))
            vobj.LineWidth = 1
        if not "OverrideUnit" in pl:
            vobj.addProperty("App::PropertyString","OverrideUnit","BuildingPart",QT_TRANSLATE_NOOP("App::Property","An optional unit to express levels"))
        if not "DisplayOffset" in pl:
            vobj.addProperty("App::PropertyPlacement","DisplayOffset","BuildingPart",QT_TRANSLATE_NOOP("App::Property","A transformation to apply to the level mark"))
            vobj.DisplayOffset = FreeCAD.Placement(FreeCAD.Vector(0,0,0),FreeCAD.Rotation(FreeCAD.Vector(1,0,0),90))
        if not "ShowLevel" in pl:
            vobj.addProperty("App::PropertyBool","ShowLevel","BuildingPart",QT_TRANSLATE_NOOP("App::Property","If true, show the level"))
            vobj.ShowLevel = True
        if not "ShowUnit" in pl:
            vobj.addProperty("App::PropertyBool","ShowUnit","BuildingPart",QT_TRANSLATE_NOOP("App::Property","If true, show the unit on the level tag"))
        if not "OriginOffset" in pl:
            vobj.addProperty("App::PropertyBool","OriginOffset","BuildingPart",QT_TRANSLATE_NOOP("App::Property","If true, display offset will affect the origin mark too"))
        if not "ShowLabel" in pl:
            vobj.addProperty("App::PropertyBool","ShowLabel","BuildingPart",QT_TRANSLATE_NOOP("App::Property","If true, the object's label is displayed"))
            vobj.ShowLabel = True
        if not "FontName" in pl:
            vobj.addProperty("App::PropertyFont","FontName","BuildingPart",QT_TRANSLATE_NOOP("App::Property","The font to be used for texts"))
            vobj.FontName = Draft.getParam("textfont","Arial")
        if not "FontSize" in pl:
            vobj.addProperty("App::PropertyLength","FontSize","BuildingPart",QT_TRANSLATE_NOOP("App::Property","The font size of texts"))
            vobj.FontSize = Draft.getParam("textheight",2.0)
        if not "DiffuseColor" in pl:
            vobj.addProperty("App::PropertyColorList","DiffuseColor","BuildingPart",QT_TRANSLATE_NOOP("App::Property","The individual face colors"))

        # Interaction properties
        if not "SetWorkingPlane" in pl:
            vobj.addProperty("App::PropertyBool","SetWorkingPlane","Interaction",QT_TRANSLATE_NOOP("App::Property","If true, when activated, the working plane will automatically adapt to this level"))
            vobj.SetWorkingPlane = True
        if not "AutoWorkingPlane" in pl:
            vobj.addProperty("App::PropertyBool","AutoWorkingPlane","Interaction",QT_TRANSLATE_NOOP("App::Property","If set to True, the working plane will be kept on Auto mode"))
        if not "ViewData" in pl:
            vobj.addProperty("App::PropertyFloatList","ViewData","Interaction",QT_TRANSLATE_NOOP("App::Property","Camera position data associated with this object"))
            vobj.setEditorMode("ViewData",2)
        if not "RestoreView" in pl:
            vobj.addProperty("App::PropertyBool","RestoreView","Interaction",QT_TRANSLATE_NOOP("App::Property","If set, the view stored in this object will be restored on double-click"))
        if not "DoubleClickActivates" in pl:
            vobj.addProperty("App::PropertyBool","DoubleClickActivates","Interaction",QT_TRANSLATE_NOOP("App::Property","If True, double-clicking this object in the tree turns it active"))

        # inventor saving
        if not "SaveInventor" in pl:
            vobj.addProperty("App::PropertyBool","SaveInventor","Interaction",QT_TRANSLATE_NOOP("App::Property","If this is enabled, the inventor representation of this object will be saved in the FreeCAD file, allowing to reference it in other files in lightweight mode."))
        if not "SavedInventor" in pl:
            vobj.addProperty("App::PropertyFileIncluded","SavedInventor","Interaction",QT_TRANSLATE_NOOP("App::Property","A slot to save the inventor representation of this object, if enabled"))
            vobj.setEditorMode("SavedInventor",2)

        # children properties
        if not "ChildrenOverride" in pl:
            vobj.addProperty("App::PropertyBool","ChildrenOverride","Children",QT_TRANSLATE_NOOP("App::Property","If true, show the objects contained in this Building Part will adopt these line, color and transparency settings"))
        if not "ChildrenLineWidth" in pl:
            vobj.addProperty("App::PropertyFloat","ChildrenLineWidth","Children",QT_TRANSLATE_NOOP("App::Property","The line width of child objects"))
            vobj.LineWidth = 1
        if not "ChildrenLineColor" in pl:
            vobj.addProperty("App::PropertyColor","ChildrenLineColor","Children",QT_TRANSLATE_NOOP("App::Property","The line color of child objects"))
            c = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/View").GetUnsigned("DefaultShapeLineColor",255)
            vobj.ChildrenLineColor = (float((c>>24)&0xFF)/255.0,float((c>>16)&0xFF)/255.0,float((c>>8)&0xFF)/255.0,0.0)
        if not "ChildrenShapeolor" in pl:
            vobj.addProperty("App::PropertyColor","ChildrenShapeColor","Children",QT_TRANSLATE_NOOP("App::Property","The shape color of child objects"))
            c = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/View").GetUnsigned("DefaultShapeColor",4294967295)
            vobj.ChildrenLineColor = (float((c>>24)&0xFF)/255.0,float((c>>16)&0xFF)/255.0,float((c>>8)&0xFF)/255.0,0.0)
        if not "ChildrenTransparency" in pl:
            vobj.addProperty("App::PropertyPercent","ChildrenTransparency","Children",QT_TRANSLATE_NOOP("App::Property","The transparency of child objects"))

        # clip properties
        if not "CutView" in pl:
            vobj.addProperty("App::PropertyBool","CutView","Clip",QT_TRANSLATE_NOOP("App::Property","Cut the view above this level"))
        if not "CutMargin" in pl:
            vobj.addProperty("App::PropertyLength","CutMargin","Clip",QT_TRANSLATE_NOOP("App::Property","The distance between the level plane and the cut line"))
            vobj.CutMargin = 1600
        if not "AutoCutView" in pl:
            vobj.addProperty("App::PropertyBool","AutoCutView","Clip",QT_TRANSLATE_NOOP("App::Property","Turn cutting on when activating this level"))

    def onDocumentRestored(self,vobj):

        selt.setProperties(vobj)

    def getIcon(self):

        import Arch_rc
        if hasattr(self,"Object"):
            if self.Object.IfcType == "Building Storey":
                return ":/icons/Arch_Floor_Tree.svg"
            elif self.Object.IfcType == "Building":
                return ":/icons/Arch_Building_Tree.svg"
        return ":/icons/Arch_BuildingPart_Tree.svg"

    def attach(self,vobj):

        self.Object = vobj.Object
        self.clip = None
        from pivy import coin
        self.sep = coin.SoGroup()
        self.mat = coin.SoMaterial()
        self.sep.addChild(self.mat)
        self.dst = coin.SoDrawStyle()
        self.sep.addChild(self.dst)
        self.lco = coin.SoCoordinate3()
        self.sep.addChild(self.lco)
        lin = coin.SoType.fromName("SoBrepEdgeSet").createInstance()
        lin.coordIndex.setValues([0,1,-1,2,3,-1,4,5,-1])
        self.sep.addChild(lin)
        self.tra = coin.SoTransform()
        self.tra.rotation.setValue(FreeCAD.Rotation(0,0,90).Q)
        self.sep.addChild(self.tra)
        self.fon = coin.SoFont()
        self.sep.addChild(self.fon)
        self.txt = coin.SoAsciiText()
        self.txt.justification = coin.SoText2.LEFT
        self.txt.string.setValue("level")
        self.sep.addChild(self.txt)
        vobj.addDisplayMode(self.sep,"Default")
        self.onChanged(vobj,"ShapeColor")
        self.onChanged(vobj,"FontName")
        self.onChanged(vobj,"ShowLevel")
        self.onChanged(vobj,"FontSize")
        return

    def getDisplayModes(self,vobj):

        return ["Default"]

    def getDefaultDisplayMode(self):

        return "Default"

    def setDisplayMode(self,mode):

        return mode

    def updateData(self,obj,prop):

        if prop in ["Placement","LevelOffset"]:
            self.onChanged(obj.ViewObject,"OverrideUnit")
        elif prop == "Shape":
            # gather all the child shapes
            colors = self.getColors(obj)
            if colors and hasattr(obj.ViewObject,"DiffuseColor"):
                if len(colors) == len(obj.Shape.Faces):
                    if colors != obj.ViewObject.DiffuseColor:
                        obj.ViewObject.DiffuseColor = colors
                        self.writeInventor(obj)
                #else:
                    #print("color mismatch:",len(colors),"colors,",len(obj.Shape.Faces),"faces")
        elif prop == "Group":
            self.onChanged(obj.ViewObject,"ChildrenOverride")
        elif prop == "Label":
            self.onChanged(obj.ViewObject,"ShowLabel")

    def getColors(self,obj):

        "recursively get the colors of objects inside this BuildingPart"

        colors = []
        for child in Draft.getGroupContents(obj):
            if hasattr(child,'Shape') and (hasattr(child.ViewObject,"DiffuseColor") or hasattr(child.ViewObject,"ShapeColor")):
                if hasattr(child.ViewObject,"DiffuseColor") and len(child.ViewObject.DiffuseColor) == len(child.Shape.Faces):
                    colors.extend(child.ViewObject.DiffuseColor)
                else:
                    c = child.ViewObject.ShapeColor[:3]+(child.ViewObject.Transparency/100.0,)
                    for i in range(len(child.Shape.Faces)):
                        colors.append(c)
        return colors

    def onChanged(self,vobj,prop):

        #print(vobj.Object.Label," - ",prop)

        if prop == "ShapeColor":
            if hasattr(vobj,"ShapeColor"):
                l = vobj.ShapeColor
                self.mat.diffuseColor.setValue([l[0],l[1],l[2]])
        elif prop == "LineWidth":
            if hasattr(vobj,"LineWidth"):
                self.dst.lineWidth = vobj.LineWidth
        elif prop == "FontName":
            if hasattr(vobj,"FontName"):
                if vobj.FontName:
                    if sys.version_info.major < 3:
                        self.fon.name = vobj.FontName.encode("utf8")
                    else:
                        self.fon.name = vobj.FontName
        elif prop in ["FontSize","DisplayOffset","OriginOffset"]:
            if hasattr(vobj,"FontSize") and hasattr(vobj,"DisplayOffset") and hasattr(vobj,"OriginOffset"):
                fs = vobj.FontSize.Value
                if fs:
                    self.fon.size = fs
                    b = vobj.DisplayOffset.Base
                    self.tra.translation.setValue([b.x+fs/8,b.y,b.z+fs/8])
                    r = vobj.DisplayOffset.Rotation
                    self.tra.rotation.setValue(r.Q)
                    if vobj.OriginOffset:
                        self.lco.point.setValues([[b.x-fs,b.y,b.z],[b.x+fs,b.y,b.z],[b.x,b.y-fs,b.z],[b.x,b.y+fs,b.z],[b.x,b.y,b.z-fs],[b.x,b.y,b.z+fs]])
                    else:
                        self.lco.point.setValues([[-fs,0,0],[fs,0,0],[0,-fs,0],[0,fs,0],[0,0,-fs],[0,0,fs]])
        elif prop in ["OverrideUnit","ShowUnit","ShowLevel","ShowLabel"]:
            if hasattr(vobj,"OverrideUnit") and hasattr(vobj,"ShowUnit") and hasattr(vobj,"ShowLevel") and hasattr(vobj,"ShowLabel"):
                z = vobj.Object.Placement.Base.z + vobj.Object.LevelOffset.Value
                q = FreeCAD.Units.Quantity(z,FreeCAD.Units.Length)
                txt = ""
                if vobj.ShowLabel:
                    txt += vobj.Object.Label
                if vobj.ShowLevel:
                    if txt:
                        txt += " "
                    if z >= 0:
                        txt += "+"
                    if vobj.OverrideUnit:
                        u = vobj.OverrideUnit
                    else:
                        u = q.getUserPreferred()[2]
                    try:
                        q = q.getValueAs(u)
                    except:
                        q = q.getValueAs(q.getUserPreferred()[2])
                    d = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Units").GetInt("Decimals",0)
                    fmt = "{0:."+ str(d) + "f}"
                    if not vobj.ShowUnit:
                        u = ""
                    txt += fmt.format(float(q)) + str(u)
                if not txt:
                    txt = " " # empty texts make coin crash...
                if isinstance(txt,unicode):
                    txt = txt.encode("utf8")
                self.txt.string.setValue(txt)
        elif prop in ["ChildrenOverride","ChildenLineWidth","ChildrenLineColor","ChildrenShapeColor","ChildrenTransparency"]:
            if hasattr(vobj,"ChildrenOverride") and vobj.ChildrenOverride:
                props = ["ChildenLineWidth","ChildrenLineColor","ChildrenShapeColor","ChildrenTransparency"]
                for child in vobj.Object.Group:
                    for prop in props:
                        if hasattr(vobj,prop) and hasattr(child.ViewObject,prop[8:]) and not hasattr(child,"ChildrenOverride"):
                            setattr(child.ViewObject,prop[8:],getattr(vobj,prop))
        elif prop in ["CutView","CutMargin"]:
            if hasattr(vobj,"CutView") and FreeCADGui.ActiveDocument.ActiveView:
                sg = FreeCADGui.ActiveDocument.ActiveView.getSceneGraph()
                if vobj.CutView:
                    from pivy import coin
                    if self.clip:
                        sg.removeChild(self.clip)
                        self.clip = None
                    for o in Draft.getGroupContents(vobj.Object.Group,walls=True):
                        if hasattr(o.ViewObject,"Lighting"):
                            o.ViewObject.Lighting = "One side"
                    self.clip = coin.SoClipPlane()
                    self.clip.on.setValue(True)
                    norm = vobj.Object.Placement.multVec(FreeCAD.Vector(0,0,1))
                    mp = vobj.Object.Placement.Base
                    mp = DraftVecUtils.project(mp,norm)
                    dist = mp.Length #- 0.1 # to not clip exactly on the section object
                    norm = norm.negative()
                    marg = 1
                    if hasattr(vobj,"CutMargin"):
                        marg = vobj.CutMargin.Value
                    if mp.getAngle(norm) > 1:
                        dist += marg
                        dist = -dist
                    else:
                        dist -= marg
                    plane = coin.SbPlane(coin.SbVec3f(norm.x,norm.y,norm.z),dist)
                    self.clip.plane.setValue(plane)
                    sg.insertChild(self.clip,0)
                else:
                    if self.clip:
                        sg.removeChild(self.clip)
                        self.clip = None
                    for o in Draft.getGroupContents(vobj.Object.Group,walls=True):
                        if hasattr(o.ViewObject,"Lighting"):
                            o.ViewObject.Lighting = "Two side"
        elif prop == "Visibility":
            # turn clipping off when turning the object off
            if hasattr(vobj,"Visibility") and not(vobj.Visibility) and hasattr(vobj,"CutView"):
                vobj.CutView = False
        elif prop == "SaveInventor":
            self.writeInventor(vobj.Object)

    def onDelete(self,vobj,subelements):

        if self.clip:
            sg.removeChild(self.clip)
            self.clip = None
        for o in Draft.getGroupContents(vobj.Object.Group,walls=True):
            if hasattr(o.ViewObject,"Lighting"):
                o.ViewObject.Lighting = "Two side"
        return True

    def doubleClicked(self,vobj):

        self.activate(vobj)
        if (not hasattr(vobj,"DoubleClickActivates")) or vobj.DoubleClickActivates:
            FreeCADGui.Selection.clearSelection()
        return True

    def activate(self,vobj):

        if FreeCADGui.ActiveDocument.ActiveView.getActiveObject("Arch") == vobj.Object:
            FreeCADGui.ActiveDocument.ActiveView.setActiveObject("Arch",None)
            if vobj.SetWorkingPlane:
                self.setWorkingPlane(restore=True)
        else:
            if (not hasattr(vobj,"DoubleClickActivates")) or vobj.DoubleClickActivates:
                FreeCADGui.ActiveDocument.ActiveView.setActiveObject("Arch",vobj.Object)
            if vobj.SetWorkingPlane:
                self.setWorkingPlane()

    def setupContextMenu(self,vobj,menu):

        from PySide import QtCore,QtGui
        import Draft_rc
        action1 = QtGui.QAction(QtGui.QIcon(":/icons/Draft_SelectPlane.svg"),"Set working plane",menu)
        QtCore.QObject.connect(action1,QtCore.SIGNAL("triggered()"),self.setWorkingPlane)
        menu.addAction(action1)
        action2 = QtGui.QAction(QtGui.QIcon(":/icons/Draft_SelectPlane.svg"),"Write camera position",menu)
        QtCore.QObject.connect(action2,QtCore.SIGNAL("triggered()"),self.writeCamera)
        menu.addAction(action2)
        action3 = QtGui.QAction(QtGui.QIcon(),"Create group...",menu)
        QtCore.QObject.connect(action3,QtCore.SIGNAL("triggered()"),self.createGroup)
        menu.addAction(action3)
        action4 = QtGui.QAction(QtGui.QIcon(),"Reorder children alphabetically",menu)
        QtCore.QObject.connect(action4,QtCore.SIGNAL("triggered()"),self.reorder)
        menu.addAction(action4)
        action5 = QtGui.QAction(QtGui.QIcon(),"Clone level up",menu)
        QtCore.QObject.connect(action5,QtCore.SIGNAL("triggered()"),self.cloneUp)
        menu.addAction(action5)

    def setWorkingPlane(self,restore=False):

        if hasattr(self,"Object") and hasattr(FreeCAD,"DraftWorkingPlane"):
            import FreeCADGui
            autoclip = False
            if hasattr(self.Object.ViewObject,"AutoCutView"):
                autoclip = self.Object.ViewObject.AutoCutView
            if restore:
                FreeCAD.DraftWorkingPlane.restore()
                if autoclip:
                    self.Object.ViewObject.CutView = False
            else:
                FreeCAD.DraftWorkingPlane.save()
                FreeCADGui.runCommand("Draft_SelectPlane")
                if autoclip:
                    self.Object.ViewObject.CutView = True
            if hasattr(FreeCADGui,"Snapper"):
                FreeCADGui.Snapper.setGrid()
            if hasattr(FreeCADGui,"draftToolBar"):
                if restore and hasattr(self,"wptext"):
                    FreeCADGui.draftToolBar.wplabel.setText(self.wptext)
                else:
                    self.wptext = FreeCADGui.draftToolBar.wplabel.text()
                    FreeCADGui.draftToolBar.wplabel.setText(self.Object.Label)
            FreeCAD.DraftWorkingPlane.lastBuildingPart = self.Object.Name

    def writeCamera(self):

        if hasattr(self,"Object"):
            from pivy import coin
            n = FreeCADGui.ActiveDocument.ActiveView.getCameraNode()
            FreeCAD.Console.PrintMessage(QT_TRANSLATE_NOOP("Draft","Writing camera position")+"\n")
            cdata = list(n.position.getValue().getValue())
            cdata.extend(list(n.orientation.getValue().getValue()))
            cdata.append(n.nearDistance.getValue())
            cdata.append(n.farDistance.getValue())
            cdata.append(n.aspectRatio.getValue())
            cdata.append(n.focalDistance.getValue())
            if isinstance(n,coin.SoOrthographicCamera):
                cdata.append(n.height.getValue())
                cdata.append(0.0) # orthograhic camera
            elif isinstance(n,coin.SoPerspectiveCamera):
                cdata.append(n.heightAngle.getValue())
                cdata.append(1.0) # perspective camera
            self.Object.ViewObject.ViewData = cdata

    def createGroup(self):

        if hasattr(self,"Object"):
            s = "FreeCAD.ActiveDocument.getObject(\"%s\").newObject(\"App::DocumentObjectGroup\",\"Group\")" % self.Object.Name
            FreeCADGui.doCommand(s)

    def reorder(self):

        if hasattr(self,"Object"):
            if hasattr(self.Object,"Group") and self.Object.Group:
                g = self.Object.Group
                g.sort(key=lambda obj: obj.Label)
                self.Object.Group = g
                FreeCAD.ActiveDocument.recompute()

    def cloneUp(self):

        if hasattr(self,"Object"):
            if not self.Object.Height.Value:
                FreeCAD.Console.PrintError("This level has no height value. Please define a height before using this function.\n")
                return
            height = self.Object.Height.Value
            ng = []
            if hasattr(self.Object,"Group") and self.Object.Group:
                for o in self.Object.Group:
                    no = Draft.clone(o)
                    Draft.move(no,FreeCAD.Vector(0,0,height))
                    ng.append(no)
            nobj = makeBuildingPart()
            Draft.formatObject(nobj,self.Object)
            nobj.Placement = self.Object.Placement
            nobj.Placement.move(FreeCAD.Vector(0,0,height))
            nobj.IfcType = self.Object.IfcType
            nobj.Height = height
            nobj.Label = self.Object.Label
            nobj.Group = ng
            for parent in self.Object.InList:
                if hasattr(parent,"Group") and hasattr(parent,"addObject") and (self.Object in parent.Group):
                    parent.addObject(nobj)
            FreeCAD.ActiveDocument.recompute()
            # fix for missing IFC attributes
            for no in ng:
                if hasattr(no,"LongName") and hasattr(no,"CloneOf") and no.CloneOf and hasattr(no.CloneOf,"LongName"):
                    no.LongName = no.CloneOf.LongName
            FreeCAD.ActiveDocument.recompute()

    def __getstate__(self):
        return None

    def __setstate__(self,state):
        return None

    def writeInventor(self,obj):

        def callback(match):
            return next(callback.v)

        if hasattr(obj.ViewObject,"SaveInventor") and obj.ViewObject.SaveInventor:
            if obj.Shape and obj.Shape.Faces and hasattr(obj,"SavedInventor"):
                colors = obj.ViewObject.DiffuseColor
                if len(colors) != len(obj.Shape.Faces):
                    print("Debug: Colors mismatch in",obj.Label)
                    colors = None
                iv = self.Object.Shape.writeInventor()
                import re
                if colors:
                    if len(re.findall("IndexedFaceSet",iv)) == len(obj.Shape.Faces):
                        # convert colors to iv representations
                        colors = ["Material { diffuseColor "+str(color[0])+" "+str(color[1])+" "+str(color[2])+"}\n    IndexedFaceSet" for color in colors]
                        # replace
                        callback.v=iter(colors)
                        iv = re.sub("IndexedFaceSet",callback,iv)
                    else:
                        print("Debug: IndexedFaceSet mismatch in",obj.Label)
                # save embedded file
                tf = tempfile.mkstemp(prefix=obj.Name,suffix=".iv")[1]
                f = open(tf,"w")
                f.write(iv)
                f.close()
                obj.SavedInventor = tf
                os.remove(tf)


if FreeCAD.GuiUp:
    FreeCADGui.addCommand('Arch_BuildingPart',CommandBuildingPart())
